#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
import inspect
from torch import optim
from ltp.core import Registrable


class Optimizer(optim.Optimizer, metaclass=Registrable):
    @classmethod
    def from_extra(cls, extra: dict, subcls=None):
        if subcls is None:
            return {'params': filter(lambda p: p.requires_grad, extra['model'].parameters())}
        sig = inspect.signature(subcls)
        params = sig.parameters.keys()
        if 'params' in params:
            return {'params': filter(lambda p: p.requires_grad, extra['model'].parameters())}
        elif 'named_params' in params:
            named_params = []
            for (n, p) in extra['model'].named_parameters():
                if p.requires_grad:
                    named_params.append((n, p))
            return {'named_params': named_params}
        else:
            return {}


Optimizer.weak_register("Adadelta", optim.Adadelta)
Optimizer.weak_register("Adagrad", optim.Adagrad)
Optimizer.weak_register("Adam", optim.Adam)
Optimizer.weak_register("AdamW", optim.AdamW)
Optimizer.weak_register("SparseAdam", optim.SparseAdam)
Optimizer.weak_register("Adamax", optim.Adamax)
Optimizer.weak_register("ASGD", optim.ASGD)
Optimizer.weak_register("SGD", optim.SGD)
Optimizer.weak_register("Rprop", optim.Rprop)
Optimizer.weak_register("RMSprop", optim.RMSprop)
Optimizer.weak_register("LBFGS", optim.LBFGS)

try:
    import torch_optimizer

    Optimizer.weak_register("AccSGD", torch_optimizer.AccSGD)
    Optimizer.weak_register("AdaBound", torch_optimizer.AdaBound)
    Optimizer.weak_register("AdaMod", torch_optimizer.AdaMod)
    Optimizer.weak_register("DiffGrad", torch_optimizer.DiffGrad)
    Optimizer.weak_register("Lamb", torch_optimizer.Lamb)
    Optimizer.weak_register("NovoGrad", torch_optimizer.NovoGrad)
    Optimizer.weak_register("PID", torch_optimizer.PID)
    Optimizer.weak_register("QHM", torch_optimizer.QHM)
    Optimizer.weak_register("RAdam", torch_optimizer.RAdam)
    Optimizer.weak_register("SGDW", torch_optimizer.SGDW)
    Optimizer.weak_register("Yogi", torch_optimizer.Yogi)
except Exception as e:
    pass

from .pretrained_optim import PretrainedOptim, BertAdamW
from .task_optim import BertAdamW4CRF

__all__ = ['Optimizer', 'PretrainedOptim', 'BertAdamW', 'BertAdamW4CRF']



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

from itertools import chain
from typing import List, Union

from ltp.core import Registrable
from ltp.data import Vocab


class Processing(object):
    """预/后处理基类"""

    def __call__(self, *args, **kwargs):
        return self.call(*args, **kwargs)

    def call(self, *args, **kwargs):
        raise NotImplementedError()


class PreProcessing(Processing, metaclass=Registrable):
    """预处理基类"""

    def call(self, x: List[str]):
        raise NotImplementedError()


class PostProcessing(Processing, metaclass=Registrable):
    """后处理基类"""

    def call(self, x, vocab: Union[Vocab, None]):
        raise NotImplementedError()


class BioEncoder(PreProcessing):
    """将词序列转换成BIO编码"""

    def call(self, x: List[str]):
        return list(chain(*(["B"] + ["I"] * (len(word) - 1) for word in x)))



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

from collections import defaultdict
import importlib
from warnings import warn
from typing import TypeVar, Type, Dict, List, Optional, Iterable, Tuple, Callable, Any
import inspect

from .exceptions import RegistrationError

T = TypeVar("T")
HookType = Callable[[Type[T], str], None]


class Registrable(type):
    """可注册元类

    给基类提供以下功能

    1. 通过名字访问子类
    2. 列出子类的类型
    3. 构造生成子类对象

    """

    def __init__(cls, name, bases, namespace, alias: str = None):
        super().__init__(name, bases, namespace)
        if not hasattr(cls, '_registry'):
            cls._registry: Dict[str, Type] = defaultdict()
            cls._hooks: Optional[List[HookType]] = None
            cls.register = classmethod(register)
            cls.weak_register = classmethod(weak_register)
            cls.hook = classmethod(hook)
            cls.by_name = classmethod(by_name)
            cls.list_available = classmethod(list_available)
            cls.is_registered = classmethod(is_registered)
            cls.iter_registered = classmethod(iter_registered)
            cls.from_params = classmethod(from_params)
            cls.default_implementation = None
        elif alias:
            cls._registry[alias] = cls
        else:
            cls._registry[name] = cls

    # Optional
    @classmethod
    def __prepare__(cls, name, bases, *, alias=False):
        return super().__prepare__(name, bases)

    # Required
    def __new__(cls, name, bases, ns, *, alias=False):
        return super().__new__(cls, name, bases, ns)


def register(
        cls: Type[T],
        name: str,
        override: bool = False,
        hooks: Optional[List[HookType]] = None,
):
    """装饰器 Class decorator for registering a subclass.

    Args:
        name: 注册名
        override (bool): 当name已经注册时，是否进行覆盖
        hooks (List[HookType]): 在注册时会被执行的Hook函数

    Raises:
        RegistrationError: 如果 override 为 false 并且 name 已经被注册
    """
    registry = cls._registry
    default_hooks = cls._hooks or []  # type: ignore

    def add_subclass_to_registry(subclass: Type[T]):
        if not inspect.isclass(subclass) or not issubclass(subclass, cls):
            raise RegistrationError(
                f"Cannot register {subclass.__name__} as {name}; "
                f"{subclass.__name__} must be a subclass of {cls.__name__}"
            )
        # Add to registry.
        # If name already registered, warn if overriding or raise an error if override not allowed.
        if name in registry:
            if not override:
                raise RegistrationError(
                    f"Cannot register {subclass.__name__} as {name}; "
                    f"name already in use for {registry[name].__name__}"
                )
            else:
                warn(f"Overriding {name} in {cls.__name__} registry")
        registry[name] = subclass
        for hook in default_hooks + (hooks or []):
            hook(subclass, name)
        return subclass

    return add_subclass_to_registry


def weak_register(
        cls: Type[T],
        name: str,
        subclass: Type[T],
        override: bool = False,
        hooks: Optional[List[HookType]] = None,
) -> Type[T]:
    """用于手动对子类进行注册

    Args:
        name (str): 子类的引用名
        subclass: 子类类型
        override(bool): 当name已经注册时，是否进行覆盖
        hooks: 在注册时会被执行的Hook函数

    Raises:
        RegistrationError: 如果 override 为 false 并且 name 已经被注册
    """
    registry = cls._registry
    default_hooks = cls._hooks or []  # type: ignore

    # 函数支持
    if not inspect.isclass(subclass) and not callable(subclass):
        raise RegistrationError(
            f"Cannot register {subclass.__name__} as {name}; "
        )
        # Add to registry.
        # If name already registered, warn if overriding or raise an error if override not allowed.
    if name in registry:
        if not override:
            raise RegistrationError(
                f"Cannot register {subclass.__name__} as {name}; "
                f"name already in use for {registry[name].__name__}"
            )
        else:
            warn(f"Overriding {name} in {cls.__name__} registry")
    registry[name] = subclass
    for hook in default_hooks + (hooks or []):
        hook(subclass, name)
    return cls


def hook(cls, hook: HookType):
    """
    函数装饰器，给某个类注册装饰器
    """
    if not cls._hooks:
        cls._hooks = []
    cls._hooks.append(hook)
    return hook


def by_name(cls: Type[T], name: str) -> Type[T]:
    """通过注册的名字取得实际的类型

    Args:
        name: 注册的名字

    Returns:
        class: Type[T] 使用 name 注册的子类

    Raises:
        RegistrationError: 如果 name 未被注册
    """
    if name in cls._registry:
        return cls._registry[name]
    elif "." in name:
        # This might be a fully qualified class name, so we'll try importing its "module"
        # and finding it there.
        parts = name.split(".")
        submodule = ".".join(parts[:-1])
        class_name = parts[-1]

        try:
            module = importlib.import_module(submodule)
        except ModuleNotFoundError:
            raise RegistrationError(
                f"tried to interpret {name} as a path to a class "
                f"but unable to import module {submodule}"
            )

        try:
            maybe_subclass = getattr(module, class_name)
        except AttributeError:
            raise RegistrationError(
                f"tried to interpret {name} as a path to a class "
                f"but unable to find class {class_name} in {submodule}"
            )

        if not inspect.isclass(maybe_subclass) or not issubclass(
                maybe_subclass, cls
        ):
            raise RegistrationError(
                f"tried to interpret {name} as a path to a class "
                f"but {class_name} is not a subclass of {cls.__name__}"
            )

        # Add subclass to registry and return it.
        cls._registry[name] = maybe_subclass
        return maybe_subclass
    else:
        # is not a qualified class name
        print(cls.list_available())
        raise RegistrationError(
            f"{name} is not a registered name for {cls.__name__}."
        )


def list_available(cls: Type[T]) -> List[str]:
    """
    列出所有的注册子类
    """
    keys: List[str]
    keys = list(cls._registry.keys())
    default = cls.default_implementation  # type: ignore

    if default is None:
        return keys
    if default not in keys:
        raise RegistrationError(
            f"Default implementation {default} is not registered"
        )
    return [default] + [k for k in keys if k != default and not k.startswith('_')]


def is_registered(cls: Type[T], name: str) -> bool:
    """
    如果 name 在类中已经注册，则返回 True
    """
    return name in cls._registry


def iter_registered(cls: Type[T]) -> Iterable[Tuple[str, Type[T]]]:
    """
    迭代已经注册的名字和对象
    """
    return cls._registry.items()


def from_params(cls, __config: dict, *args, extra: dict = None, **kwargs):
    """使用 config 生成对象

    Args:
        cls: 类型
        __config: 配置项，通常为字典，形如 {'class':'ClassName', 'init':{ 'arg': arg } }
        *args: 直接传入的arg
        extra: 根据需要传入的数据
        **kwargs: 其他参数

    Returns:
        根据参数生成的对象
    """
    if "class" not in __config:
        raise RegistrationError("Not available config!!!")

    # todo Generate Config Report
    # arg_names = inspect.getfullargspec(cls)

    class_ = cls.by_name(__config["class"])

    if "init" in __config:
        kwargs.update(__config['init'])  # 以配置文件为主

    if inspect.ismethod(getattr(class_, 'from_extra', None)):
        if extra is None:
            extra = {'config': __config}
        else:
            extra['config'] = __config
        kwargs.update(class_.from_extra(extra=extra))
    elif inspect.ismethod(getattr(cls, 'from_extra', None)):
        if extra is None:
            extra = {'config': __config}
        else:
            extra['config'] = __config
        kwargs.update(cls.from_extra(extra=extra, subcls=class_))

    return class_(*args, **kwargs)



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
from . import Callback


class MetricCallback(Callback, alias="_metrics_wrapper"):
    def __init__(self, metrics):
        super(MetricCallback, self).__init__(1, None)
        self.metrics = metrics
        self.last_epoch = 0

    def init(self, executor):
        for metric in self.metrics:
            for name, value in metric.defaults.items():
                executor.trainer.state.add_attribute(name, value)
                executor.add_progressbar_metric(name)

    def call(self, executor):
        if self.last_epoch != executor.trainer.state.current_epoch:
            for metric in self.metrics:
                metric.clear()

        self.last_epoch = executor.trainer.state.current_epoch
        for metric in self.metrics:
            metric.step(executor.trainer.state.last_y_pred, executor.trainer.state.last_y)
            for attr, value in metric.compute().items():
                executor.trainer.state.update_attribute(attr, value)



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
import torch
from . import Predictor
from ltp.utils.seqeval import get_entities
from transformers import PreTrainedTokenizer


class Segment(Predictor, alias='segment'):
    def __init__(self, fields, text_field='text'):
        super().__init__(fields)
        self.text_field = text_field
        tokenizer = self.fields[text_field].tokenizer
        tokenizer: PreTrainedTokenizer

        self.id2text = tokenizer
        self.id2label = ['B-W', 'I-W']

    def convert_idx_to_name(self, y, array_len, id2label):
        return [[id2label[idx] for idx in row[:row_len]] for row, row_len in zip(y, array_len)]

    def predict(self, inputs, pred):
        input_text = getattr(inputs, self.text_field)
        target_len = input_text['text_length'].cpu().detach().numpy().tolist()

        pred = torch.argmax(pred, dim=-1).cpu().detach().numpy()
        text = input_text['input_ids'].cpu().detach().numpy()
        pred = self.convert_idx_to_name(pred, target_len, id2label=self.id2label)
        res = []
        for text, pred_single, length in zip(text, pred, target_len):
            text = [self.id2text.convert_ids_to_tokens(int(char)) for char in text[1:1 + length]]
            entities = get_entities(pred_single)
            sentence = ["".join(text[entity[1]:entity[2] + 1]) for entity in entities]
            res.append(sentence)
        return res



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
from torch import Tensor
from ltp.nn import MLP
from ltp import nn

from . import Module


class BiLinearCRF(Module):
    def __init__(self, input_size, label_num, dropout: float = 0.2, hidden_size=None, **kwargs):
        super().__init__(input_size, label_num, dropout)
        activation = kwargs.pop('activation', {'LeakyReLU': {}})
        self.mlp_rel_h = MLP(input_size, hidden_size, dropout=dropout, **activation)
        self.mlp_rel_d = MLP(input_size, hidden_size, dropout=dropout, **activation)
        self.biaffine = nn.Bilinear(hidden_size, hidden_size, label_num, bias_x=True, bias_y=True, expand=True)

        self.crf = nn.CRF(label_num)

    def forward(self, inputs: Tensor, length: Tensor, gold=None):
        rel_h = self.mlp_rel_h(inputs)
        rel_d = self.mlp_rel_d(inputs)

        logits = self.biaffine(rel_h, rel_d).permute(0, 2, 3, 1)
        return logits, length, self.crf



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

from collections import Counter, OrderedDict
from itertools import chain
import torch

from torchtext.vocab import Vocab

from . import Field
from ltp.data.dataset import Dataset
from ltp.const import UNK


class LabelField(Field, alias='label'):
    """
    可以用于文本分类等领域
    """
    vocab_cls = Vocab

    dtypes = {
        torch.float32: float,
        torch.float: float,
        torch.float64: float,
        torch.double: float,
        torch.float16: float,
        torch.half: float,

        torch.uint8: int,
        torch.int8: int,
        torch.int16: int,
        torch.short: int,
        torch.int32: int,
        torch.int: int,
        torch.int64: int,
        torch.long: int,
    }

    ignore = ['dtype']

    def __init__(self, name, unk: str = UNK, preprocessing=None, postprocessing=None, use_vocab=True, dtype=torch.long,
                 is_target: bool = False):
        super(LabelField, self).__init__(name, preprocessing, postprocessing, is_target)

        self.unk = unk
        self.dtype = dtype
        self.use_vocab = use_vocab

    def build_vocab(self, *args, **kwargs):
        counter = Counter()
        sources = []
        for arg in args:
            if isinstance(arg, Dataset):
                sources += [getattr(arg, name) for name, field in
                            arg.fields.items() if field is self]
            else:
                sources.append(arg)
        for data in sources:
            for x in data:
                try:
                    counter.update(x)
                except TypeError:
                    counter.update(chain.from_iterable(x))
        specials = list(
            OrderedDict.fromkeys(
                tok for tok in [self.unk] + kwargs.pop('specials', []) if tok is not None)
        )
        self.vocab = self.vocab_cls(counter, specials=specials, **kwargs)

    def numericalize(self, arr, device=None):
        if self.use_vocab:
            arr = [[self.vocab.stoi[x] for x in ex] for ex in arr]
            if self.postprocessing is not None:
                arr = self.postprocessing(arr, self.vocab)
        else:
            if self.dtype not in self.dtypes:
                raise ValueError(
                    f"Specified Field dtype {self.dtype} can not be used with "
                    "use_vocab=False because we do not know how to numericalize it. "
                )
            numericalization_func = self.dtypes[self.dtype]
            arr = [numericalization_func(x) for x in arr]
            if self.postprocessing is not None:
                arr = self.postprocessing(arr, None)

        var = torch.tensor(arr, dtype=self.dtype, device=device)
        var = var.contiguous()
        return var

    def preprocess(self, x):
        if self.preprocessing is not None:
            return self.preprocessing(x)
        else:
            return x

    def process(self, batch, device=None):
        tensor = self.numericalize(batch, device=device)
        return tensor



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
import os

import numpy
from tqdm import tqdm

from . import Dataset
from ltp.data.example import Example
from ltp.data.utils import iter_blocks
from ltp.utils import expand_bio


class CorpusDataset(Dataset, alias="Corpus"):
    """
    Dataset 注册名: Corpus
    Conll 文本标准，不同Field之间使用 Tab 分割，不同句子之间有一空行，例子如下::

        科学	_	O	B-ARG0	B-ARG0	O
        发展观	_	O	I-ARG0	I-ARG0	O
        绝对	_	O	B-ARGM-ADV	B-ARGM-ADV	O
        不	_	O	B-ARGM-ADV	B-ARGM-ADV	O
    """

    def __init__(self, path, file, fields, multi_field=None,
                 split=None, strip=None, proxy_property=None, **kwargs):
        filename = os.path.join(path, file)
        examples = list(self.iter(filename, fields, multi_field, split, strip, proxy_property))
        super(CorpusDataset, self).__init__(examples, fields, **kwargs)

    @staticmethod
    def build_slice(fields, multi_field=None, ignore_fields=None):
        used_fields = []
        field_row = []

        ignored = 0
        if ignore_fields is None:
            ignore_fields = {}

        for idx, field in enumerate(fields):
            if field is None:
                continue
            used_fields.append(field)
            if field[0] in ignore_fields:
                ignored += 1
                field_row.append(-1)
                continue
            field_row.append(idx - ignored)
        if multi_field is None:
            return used_fields, field_row

        field_names = [field[0] if field is not None else str(idx) for idx, field in enumerate(fields)]
        if multi_field not in field_names:
            return used_fields, field_row

        mf_tag_idx = field_names.index(multi_field)
        mf_tag_id = field_row[mf_tag_idx]
        if mf_tag_idx == len(field_names) - len(ignore_fields) - 1:  # 是最后一个
            field_slices = field_row[:mf_tag_idx] + [slice(mf_tag_id, None)]
        else:
            raise NotImplementedError("Multifield must be last row")

        return used_fields, [None if field_name in ignore_fields else field_slices[idx]
                             for idx, (field_name, field) in enumerate(used_fields)]

    def iter(self, filename: str, fields, multi_field: str = None, split=None, strip=None, proxy_property: dict = None):
        fields, fields_slices = self.build_slice(fields, multi_field, proxy_property)
        if proxy_property is not None:
            field_map = {field[0]: idx for idx, field in enumerate(fields)}
            for proxy, source in proxy_property.items():
                if proxy in field_map:
                    fields_slices[field_map[proxy]] = fields_slices[field_map[source]]

        for line_num, block in tqdm(list(iter_blocks(filename, split, strip))):
            values = [list(value) for value in zip(*block)]
            values = [values[field_slice] for field_slice in fields_slices]
            processed, more = self.post_fn(values)

            if more:
                for values in processed:
                    try:
                        yield Example.fromlist(values, fields)
                    except Exception as e:
                        print(line_num, e)
            else:
                try:
                    yield Example.fromlist(processed, fields)
                except Exception as e:
                    print(line_num, e)

    def post_fn(self, input):
        return input, False



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
from ltp.core import Registrable


class State(object):
    global_step = 0
    current_epoch = 0
    last_x = None
    last_y = None
    last_y_pred = None
    last_train_loss = float("inf")

    def get(self, attribute_name: str):
        return getattr(self, attribute_name)

    def load_state_dict(self, state):
        for k, v in state.__dict__.items():
            setattr(self, k, v)
        return self

    def add_attribute(self, attribute, value):
        if not hasattr(self, attribute):
            setattr(self, attribute, value)
        return self

    def update_attribute(self, attribute, value):
        if hasattr(self, attribute):
            setattr(self, attribute, value)
        return self

    def reset(self):
        self.global_step = 0
        self.current_epoch = 0
        self.last_x = None
        self.last_y = None
        self.last_y_pred = None
        self.last_train_loss = float("inf")
        return self

    def __eq__(self, other):
        if not isinstance(other, State):
            return NotImplemented

        return self.current_epoch == other.current_epoch \
               and self.global_step == other.global_step \
               and self.last_x == other.last_x \
               and self.last_y == other.last_y \
               and self.last_y_pred == other.last_y_pred \
               and self.last_train_loss == other.last_train_loss


class Trainer(metaclass=Registrable):
    state: State

    def __init__(self, config):
        self.state = State()
        self.config = config
        self.total_steps = 0
        self.predictor = None

    def init(self, total_steps):
        self.total_steps = total_steps

    def before_train(self):
        """
        每个EPOCH之前执行
        """
        pass

    def train(self, batch, task):
        raise NotImplementedError()

    def after_train(self):
        """
        每个EPOCH之后执行
        """
        pass

    def before_eval(self, task):
        """
        Eval 之前执行
        """
        pass

    def eval(self, batch, task):
        raise NotImplementedError()

    def after_eval(self, task):
        """
        Eval 之后执行
        """
        pass

    def before_predict(self, task):
        """
        预测之前执行
        """
        pass

    def predict(self, batch, task):
        raise NotImplementedError()

    def after_predict(self, task):
        """
        预测之后执行
        """
        pass


from .common_trainer import CommonTrainer
from .multi_task_distiller import MultiTaskDistiller

__all__ = ['Trainer', 'CommonTrainer', 'MultiTaskDistiller']



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
from typing import List

import torch
from torch import Tensor
from .eisner import eisner
from .initial import initial_parameter
from .clip_grad_norm import clip_grad_norm
from .deprecated import deprecated, deprecated_param
from .sent_split import split_sentence
from .ltp_trie import Trie


def cycle(iterable):
    while True:
        for x in iterable:
            yield x


def length_to_mask(length: Tensor, max_len: int = None, dtype=None):
    """
    将 Sequence length 转换成 Mask

    >>> lens = [3, 5, 4]
    >>> length_to_mask(length)
    >>> [[1, 1, 1, 0, 0],\
        [1, 1, 1, 1, 1], \
        [1, 1, 1, 1, 0]]

    :param length: [batch,]
    :param max_len: 最大长度
    :param dtype: nn.dtype
    :return: batch * max_len : 如果 max_len is None
    :return: batch * max(length) : 如果 max_len is None
    """
    assert len(length.shape) == 1, 'Length shape should be 1 dimensional.'
    if max_len is None:
        max_len = max_len or torch.max(length)
    mask = torch.arange(max_len, device=length.device, dtype=length.dtype). \
               expand(length.shape[0], max_len) < length.unsqueeze(1)
    if dtype is not None:
        mask = mask.to(dtype)
    return mask


def select_logits_with_mask(logits, mask):
    if len(logits.shape) == 3:
        mask = mask.unsqueeze(-1).expand_as(logits).to(torch.bool)
        logits_select = torch.masked_select(logits, mask).view(-1, logits.size(-1))
    else:
        logits_select = logits  # Logits_mask has no effect on logits of shape (batch_size, logits_to_be_softmaxed)
    return logits_select


def expand_bio(word_lens: List[int], srl: List[str]):
    res = []
    for word_len, tag in zip(word_lens, srl):
        if tag.startswith('B-'):
            res.extend([tag] + ['I-' + tag[2:]] * (word_len - 1))
        else:
            res.extend([tag] * word_len)
    return res


def pad_sequence(sequences, batch_first=True, pad_value=0):
    def length(sequence):
        if isinstance(sequence, torch.Tensor):
            return sequence.size(0)
        if isinstance(sequence, list):
            return len(sequence)

    lengths, sequences = zip(*[(length(sequence), torch.as_tensor(sequence)) for sequence in sequences])

    return torch.as_tensor(lengths), torch.nn.utils.rnn.pad_sequence(
        sequences, batch_first=batch_first, padding_value=pad_value)


try:
    from ltp_plugin import get_entities, is_chinese_char, segment_decode

    USE_PLUGIN = False

except Exception as e:
    from .seqeval import get_entities

    USE_PLUGIN = False


    def segment_decode(inputs, segment_output, offsets, words):
        pass


    def is_chinese_char(cp):
        """Checks whether CP is the codepoint of a CJK character."""
        # This defines a "chinese character" as anything in the CJK Unicode block:
        #   https://en.wikipedia.org/wiki/CJK_Unified_Ideographs_(Unicode_block)
        #
        # Note that the CJK Unicode block is NOT all Japanese and Korean characters,
        # despite its name. The modern Korean Hangul alphabet is a different block,
        # as is Japanese Hiragana and Katakana. Those alphabets are used to write
        # space-separated words, so they are not treated specially and handled
        # like the all of the other languages.
        cp = ord(cp)
        if (
                (cp >= 0x4E00 and cp <= 0x9FFF)
                or (cp >= 0x3400 and cp <= 0x4DBF)  #
                or (cp >= 0x20000 and cp <= 0x2A6DF)  #
                or (cp >= 0x2A700 and cp <= 0x2B73F)  #
                or (cp >= 0x2B740 and cp <= 0x2B81F)  #
                or (cp >= 0x2B820 and cp <= 0x2CEAF)  #
                or (cp >= 0xF900 and cp <= 0xFAFF)
                or (cp >= 0x2F800 and cp <= 0x2FA1F)  #
        ):  #
            return True

        return False



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
import torch, math
from torch import nn, Tensor
from torch.nn import functional as F


# Change from Pytorch nn.Bilinear
class Bilinear(nn.Module):
    __constants__ = ['in1_features', 'in2_features', 'out_features', 'bias_x', 'bias_y']

    def __init__(self, in1_features, in2_features, out_features, expand=False, bias_x=True, bias_y=True):
        super(Bilinear, self).__init__()
        self.expand = expand
        self.bias_x = bias_x
        self.bias_y = bias_y
        self.in1_features = in1_features
        self.in2_features = in2_features
        self.out_features = out_features
        self.weight = nn.Parameter(
            torch.zeros(out_features, in1_features + bias_x, in2_features + bias_y),
            requires_grad=True
        )
        self.reset_parameters()

    def reset_parameters(self):
        bound = 1 / math.sqrt(self.weight.size(1))
        nn.init.uniform_(self.weight, -bound, bound)

    def forward(self, x1: Tensor, x2: Tensor):
        if self.bias_x:
            x1 = torch.cat((x1, torch.ones_like(x1[..., :1])), -1)
        if self.bias_y:
            x2 = torch.cat((x2, torch.ones_like(x2[..., :1])), -1)
        if self.expand:
            # [batch_size, n_out, seq_len, seq_len]
            s = torch.einsum('bxi,oij,byj->boxy', x1, self.weight, x2)
            return s
        # [batch_size, n_out, seq_len]
        return F.bilinear(x1, x2, self.weight, None)

    def extra_repr(self):
        return 'in1_features={}, in2_features={}, out_features={}, bias_x={}, bias_y={}'.format(
            self.in1_features, self.in2_features, self.out_features, self.bias_x, self.bias_y
        )



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
import os
from . import Callback

default_main_tag = "tensorboard"
default_save_directory = './log'


class TensorboardCallback(Callback, alias="tensorboard"):
    def __init__(self, iteration, epoch, save_directory=default_save_directory, main_tag=default_main_tag,
                 data_extractor=lambda state: {"loss": state.last_train_loss}):
        super(TensorboardCallback, self).__init__(iteration, epoch)
        from torch.utils.tensorboard import SummaryWriter

        self.main_tag = main_tag
        self.save_directory = save_directory
        self.data_extractor = data_extractor
        self.writer = SummaryWriter(self.save_directory)

        os.makedirs(save_directory, exist_ok=True)

    def call(self, executor):
        self.__save(executor.trainer.state)

    def __save(self, executor_state):
        self.writer.add_scalars(
            self.main_tag,
            self.data_extractor(executor_state),
            global_step=executor_state.global_step
        )



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
import torch
from torch import Tensor
from torchtext.data import Batch

from . import Trainer
from ltp.utils import clip_grad_norm
from ltp.predict import Predictor
from ltp.train.loss import Loss


class CommonTrainer(Trainer, alias='default'):
    def __init__(self, config):
        super().__init__(config)
        self.verbose = self.config.verbose
        self.task = self.config.tasks['default']
        self.model = self.task.model

        if hasattr(self.model, 'extractor'):
            self.data_extractor = lambda batch: self.model.extractor(batch)
        else:
            self.data_extractor = self.__data_extractor
        if self.verbose:
            print(self.model)

    def __data_extractor(self, batch: Batch):
        x = tuple(getattr(batch, name) for name, _ in self.task.fields if name in batch.input_fields)
        y = tuple(getattr(batch, name) for name, _ in self.task.fields if name in batch.target_fields)
        return x, y

    def before_train(self):
        self.model.train()

    def train(self, batch, task):
        self.task.optimizer.zero_grad()
        x, y = self.data_extractor(batch)

        if len(y) == 0:
            y = x
        if len(y) == 1:
            y = y[0]

        if isinstance(x, tuple):
            y_pred = self.model(*x, gold=y)
        elif isinstance(x, dict):
            y_pred = self.model(**x, gold=y)
        else:
            raise NotImplementedError("Model.extrator 的返回值 x 必须是一个tuple或者一个 dict")

        if isinstance(self.task.loss, Loss):
            loss = self.task.loss(y_pred, y)
        elif isinstance(y, Tensor):
            shape = y_pred.shape[-1]
            loss = self.task.loss(y_pred.contiguous().view((-1, shape)), y.contiguous().view(-1))
        else:
            raise NotImplementedError("多输出结果需要自行实现Loss")
        loss_item = loss.item()
        loss.backward()

        clip_grad_norm(
            self.model.named_parameters() if self.config.pretrained_grad_norm else self.model.parameters(),
            self.config.grad_norm, norm_type=self.config.norm_type, pretrained_norm=self.config.pretrained_grad_norm
        )
        self.task.optimizer.step()
        if self.task.scheduler_type == 'step' and self.task.scheduler:
            self.task.scheduler.step()
        return loss_item

    def after_train(self):
        if self.task.scheduler_type == 'epoch' and self.task.scheduler:
            self.task.scheduler.step()
        self.task.save(self.state)

    def before_eval(self, task):
        self.previous_training_flag = self.model.training
        self.model.eval()

    def eval(self, batch, task):
        self.model.eval()
        x, y = self.data_extractor(batch)

        if len(y) == 0:
            y = x
        if len(y) == 1:
            y = y[0]

        if isinstance(x, tuple):
            y_pred = self.model(*x)
        elif isinstance(x, dict):
            y_pred = self.model(**x)
        else:
            raise NotImplementedError("Model.extrator 的返回值 x 必须是一个tuple或者一个 dict")

        return x, y_pred, y

    def after_eval(self, task):
        self.model.train(self.previous_training_flag)

    def before_predict(self, task):
        self.previous_training_flag = self.model.training
        self.model.eval()
        if self.task.config.predictor:
            self.predictor = Predictor.from_params(
                self.task.config.predictor,
                fields=self.task.fields
            )
        else:
            print("No Predictor")

    def predict(self, batch, task):
        self.model.eval()
        x, y = self.data_extractor(batch)

        if isinstance(x, tuple):
            y_pred = self.model(*x)
        elif isinstance(x, dict):
            y_pred = self.model(**x)
        else:
            raise NotImplementedError("Model.extrator 的返回值 x 必须是一个tuple或者一个 dict")

        return self.predictor(batch, y_pred)

    def after_predict(self, task):
        self.model.train(self.previous_training_flag)



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
from typing import Dict
import torch
import toml

from ltp.exe.task import TaskConfig, Task


class Config(object):
    """
    全局配置文件
    """
    tasks: Dict[str, Task]
    temperature_scheduler: dict
    tau: float

    def __init__(self, config_path, device=None):
        config = toml.load(config_path)
        if device is None:
            self.device = torch.device('cuda', index=0) if torch.cuda.is_available() else torch.device('cpu')
        else:
            self.device = torch.device(device)

        global_config = config.pop("Global", {})
        self.verbose = global_config.pop("verbose", False)
        self.tau = global_config.pop("tau", 1.0)
        self.batch_size = global_config.pop("batch_size", 5)
        self.torch_seed = global_config.pop('seed', None)
        self.epoch = global_config.pop("epoch", 10)
        self.early_stopping = global_config.pop('early_stopping', None)
        self.epoch_size = global_config.pop('epoch_size', None)

        self.grad_norm = global_config.pop("grad_norm", None)
        self.norm_type = global_config.pop("norm_type", None)
        self.pretrained_grad_norm = global_config.pop("pretrained_grad_norm", None)

        self.callbacks = config.pop("Callbacks", [])

        self.tasks = {}
        # task special
        tasks = config.pop("Task", None)
        self.tasks['default'] = Task(TaskConfig(config, self), self.device)
        if tasks:
            for task in tasks:
                self.tasks[task['name']] = Task(TaskConfig(toml.load(task['config']), self), self.device)

        self.trainer = config.pop(
            'Trainer', {'class': 'default'} if len(self.tasks) == 1 else {'class': 'multi'}
        )

        # 其他的配置都是任务特有的
        for key, value in config.items():
            setattr(self, self._camel_to_underline(key), value)

    def _camel_to_underline(self, camel_format):
        """
        驼峰命名格式转下划线命名格式
        """
        lst = []
        for index, char in enumerate(camel_format):
            if char.isupper() and index != 0:
                lst.append("_")
            lst.append(char)

        return "".join(lst).lower()



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
from typing import Optional
import torch, math
from torch import Tensor, nn
import torch.nn.functional as F
from ltp.utils import length_to_mask
from ltp.nn import ffnn


class RelativeEmbedding(nn.Module):
    """This module produces sinusoidal positional embeddings of any length.
    Padding symbols are ignored.
    """

    def __init__(self, embedding_dim, padding_idx, init_size=1024):
        """
        :param embedding_dim: 每个位置的dimension
        :param padding_idx:
        :param init_size:
        """
        super().__init__()
        self.embedding_dim = embedding_dim
        self.padding_idx = padding_idx
        assert init_size % 2 == 0
        weights = self.get_embedding(
            init_size + 1,
            embedding_dim,
            padding_idx,
        )
        self.register_buffer('weights', weights)
        self.register_buffer('_float_tensor', torch.as_tensor(1.0))

    def get_embedding(self, num_embeddings, embedding_dim, padding_idx=None):
        """Build sinusoidal embeddings.
        This matches the implementation in tensor2tensor, but differs slightly
        from the description in Section 3.5 of "Attention Is All You Need".
        """
        half_dim = embedding_dim // 2
        emb = math.log(10000) / (half_dim - 1)
        emb = torch.exp(torch.arange(half_dim, dtype=torch.float) * -emb)
        emb = torch.arange(-num_embeddings // 2, num_embeddings // 2, dtype=torch.float).unsqueeze(1) * emb.unsqueeze(0)
        emb = torch.cat([torch.sin(emb), torch.cos(emb)], dim=1).view(num_embeddings, -1)
        if embedding_dim % 2 == 1:
            # zero pad
            emb = torch.cat([emb, torch.zeros(num_embeddings, 1)], dim=1)
        if padding_idx is not None:
            emb[padding_idx, :] = 0
        self.origin_shift = num_embeddings // 2 + 1
        return emb

    def forward(self, inputs: Tensor):
        """Input is expected to be of size [bsz x seqlen].
        """
        bsz, seq_len = inputs.size()
        max_pos = self.padding_idx + seq_len
        if max_pos > self.origin_shift:
            # recompute/expand embeddings if needed
            weights = self.get_embedding(
                max_pos * 2,
                self.embedding_dim,
                self.padding_idx,
            )
            weights = weights.to(self._float_tensor)
            del self.weights
            self.origin_shift = weights.size(0) // 2
            self.register_buffer('weights', weights)

        positions = torch.arange(-seq_len, seq_len).to(inputs.device).long() + self.origin_shift  # 2*seq_len
        embed = self.weights.index_select(0, positions.long()).detach()
        return embed


class RelativeMultiHeadAttn(nn.Module):
    def __init__(self, input_size, num_head, dropout, r_w_bias=None, r_r_bias=None, max_length=1024):
        """
        :param int input_size:
        :param int num_head:
        :param dropout: 对attention map的dropout
        :param r_w_bias: n_head x head_dim or None
        :param r_r_bias: n_head x head_dim or None
        """
        super().__init__()
        self.qv_linear = nn.Linear(input_size, input_size * 2, bias=False)
        self.n_head = num_head
        self.head_dim = input_size // num_head
        self.dropout_layer = nn.Dropout(dropout)

        self.pos_embed = RelativeEmbedding(input_size // num_head, 0, max_length)

        self.scale = 1

        if r_r_bias is None or r_w_bias is None:  # Biases are not shared
            self.r_r_bias = nn.Parameter(nn.init.xavier_normal_(torch.zeros(num_head, input_size // num_head)))
            self.r_w_bias = nn.Parameter(nn.init.xavier_normal_(torch.zeros(num_head, input_size // num_head)))
        else:
            self.r_r_bias = r_r_bias  # r_r_bias就是v
            self.r_w_bias = r_w_bias  # r_w_bias就是u

    def forward(self, x, mask):
        """
        :param x: batch_size x max_len x d_model
        :param mask: batch_size x max_len
        :return:
        """

        batch_size, max_len, d_model = x.size()
        pos_embed = self.pos_embed(mask)  # l x head_dim

        qv = self.qv_linear(x)  # batch_size x max_len x d_model2
        q, v = torch.chunk(qv, chunks=2, dim=-1)
        q = q.view(batch_size, max_len, self.n_head, -1).transpose(1, 2)
        k = x.view(batch_size, max_len, self.n_head, -1).transpose(1, 2)
        v = v.view(batch_size, max_len, self.n_head, -1).transpose(1, 2)  # b x n x l x d

        rw_head_q = q + self.r_r_bias[:, None]
        AC = torch.einsum('bnqd,bnkd->bnqk', rw_head_q, k)  # b x n x l x d, n是head

        D_ = torch.einsum('nd,ld->nl', self.r_w_bias, pos_embed)[None, :, None]  # head x 2max_len, 每个head对位置的bias
        B_ = torch.einsum('bnqd,ld->bnql', q, pos_embed)  # bsz x head  x max_len x 2max_len，每个query对每个shift的偏移
        BD = B_ + D_  # bsz x head x max_len x 2max_len, 要转换为bsz x head x max_len x max_len
        BD = self._shift(BD)
        attn = AC + BD

        attn = attn.masked_fill(mask[:, None, None, :].eq(0), float('-inf'))

        attn = F.softmax(attn, dim=-1)
        attn = self.dropout_layer(attn)
        v = torch.matmul(attn, v).transpose(1, 2).reshape(batch_size, max_len, d_model)  # b x n x l x d

        return v

    def _shift(self, BD):
        """
        类似
        -3 -2 -1 0 1 2
        -3 -2 -1 0 1 2
        -3 -2 -1 0 1 2
        转换为
        0   1  2
        -1  0  1
        -2 -1  0
        :param BD: batch_size x n_head x max_len x 2max_len
        :return: batch_size x n_head x max_len x max_len
        """
        bsz, n_head, max_len, _ = BD.size()
        zero_pad = BD.new_zeros(bsz, n_head, max_len, 1)
        BD = torch.cat([BD, zero_pad], dim=-1).view(bsz, n_head, -1, max_len)  # bsz x n_head x (2max_len+1) x max_len
        BD = BD[:, :, :-1].view(bsz, n_head, max_len, -1)  # bsz x n_head x 2max_len x max_len
        BD = BD[:, :, :, max_len:]
        return BD


class RelativeTransformerLayer(nn.Module):
    def __init__(self, input_size, hidden_size, num_heads, dropout=0.2, after_norm=True, max_length=1024):
        """
        :param int input_size: 一般512之类的
        :param int hidden_size: FFN中间层的dimension的大小\
        :param num_heads: self attention模块
        :param bool after_norm: norm的位置不一样，如果为False，则embedding可以直接连到输出
        :param float dropout: 一共三个位置的dropout的大小
        """
        super().__init__()

        self.after_norm = after_norm
        self.norm1 = nn.LayerNorm(input_size)
        self.norm2 = nn.LayerNorm(input_size)
        self.self_attn = RelativeMultiHeadAttn(input_size, num_heads, dropout=dropout, max_length=max_length)
        self.ffn = ffnn(input_size, input_size, hidden_size, dropout, depth=2, last_dropout=True)

    def forward(self, x, mask):
        """
        :param x: batch_size x max_len x hidden_size
        :param mask: batch_size x max_len, 为0的地方为pad
        :return: batch_size x max_len x hidden_size
        """
        residual = x
        if not self.after_norm:
            x = self.norm1(x)

        x = self.self_attn(x, mask)
        x = x + residual
        if self.after_norm:
            x = self.norm1(x)
        residual = x
        if not self.after_norm:
            x = self.norm2(x)
        x = self.ffn(x)
        x = residual + x
        if self.after_norm:
            x = self.norm2(x)
        return x


class RelativeTransformer(nn.Module):
    def __init__(self, input_size, num_layers, hidden_size, num_heads, dropout, after_norm=True, max_length=1024):
        super().__init__()
        self.layers = nn.ModuleList([
            RelativeTransformerLayer(input_size, hidden_size, num_heads, dropout, after_norm, max_length=max_length)
            for _ in range(num_layers)
        ])

    def forward(self, x: Tensor, length: Tensor, gold: Optional = None):
        """
        :param x: batch_size x max_len
        :param length: sequence length, B
        """

        mask = length_to_mask(length, dtype=torch.long)

        for layer in self.layers:
            x = layer(x, mask)
        return x, length, gold



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
# REF: https://github.com/yzhangcs/parser/blob/master/parser/utils/alg.py

import torch
from torch.nn.utils.rnn import pad_sequence


def eisner(scores, mask) -> torch.Tensor:
    lens = mask.sum(1)
    batch_size, seq_len, _ = scores.shape
    scores = scores.permute(2, 1, 0)
    s_i = torch.full_like(scores, float('-inf'))
    s_c = torch.full_like(scores, float('-inf'))
    p_i = scores.new_zeros(seq_len, seq_len, batch_size).long()
    p_c = scores.new_zeros(seq_len, seq_len, batch_size).long()
    s_c.diagonal().fill_(0)

    for w in range(1, seq_len):
        n = seq_len - w
        starts = p_i.new_tensor(range(n)).unsqueeze(0)
        # ilr = C(i->r) + C(j->r+1)
        ilr = stripe(s_c, n, w) + stripe(s_c, n, w, (w, 1))
        # [batch_size, n, w]
        ilr = ilr.permute(2, 0, 1)
        il = ilr + scores.diagonal(-w).unsqueeze(-1)
        # I(j->i) = max(C(i->r) + C(j->r+1) + s(j->i)), i <= r < j
        il_span, il_path = il.max(-1)
        s_i.diagonal(-w).copy_(il_span)
        p_i.diagonal(-w).copy_(il_path + starts)
        ir = ilr + scores.diagonal(w).unsqueeze(-1)
        # I(i->j) = max(C(i->r) + C(j->r+1) + s(i->j)), i <= r < j
        ir_span, ir_path = ir.max(-1)
        s_i.diagonal(w).copy_(ir_span)
        p_i.diagonal(w).copy_(ir_path + starts)

        # C(j->i) = max(C(r->i) + I(j->r)), i <= r < j
        cl = stripe(s_c, n, w, (0, 0), 0) + stripe(s_i, n, w, (w, 0))
        cl_span, cl_path = cl.permute(2, 0, 1).max(-1)
        s_c.diagonal(-w).copy_(cl_span)
        p_c.diagonal(-w).copy_(cl_path + starts)
        # C(i->j) = max(I(i->r) + C(r->j)), i < r <= j
        cr = stripe(s_i, n, w, (0, 1)) + stripe(s_c, n, w, (1, w), 0)
        cr_span, cr_path = cr.permute(2, 0, 1).max(-1)
        s_c.diagonal(w).copy_(cr_span)
        s_c[0, w][lens.ne(w)] = float('-inf')
        p_c.diagonal(w).copy_(cr_path + starts + 1)

    predicts = []
    p_c = p_c.permute(2, 0, 1).cpu()
    p_i = p_i.permute(2, 0, 1).cpu()
    for i, length in enumerate(lens.tolist()):
        heads = p_c.new_ones(length + 1, dtype=torch.long)
        backtrack(p_i[i], p_c[i], heads, 0, length, True)
        predicts.append(heads.to(mask.device))

    return pad_sequence(predicts, True).to(scores.device)


def backtrack(p_i, p_c, heads, i, j, complete):
    if i == j:
        return
    if complete:
        r = p_c[i, j]
        backtrack(p_i, p_c, heads, i, r, False)
        backtrack(p_i, p_c, heads, r, j, True)
    else:
        r, heads[j] = p_i[i, j], i
        i, j = sorted((i, j))
        backtrack(p_i, p_c, heads, i, r, True)
        backtrack(p_i, p_c, heads, j, r + 1, True)


def stripe(x, n, w, offset=(0, 0), dim=1):
    r'''Returns a diagonal stripe of the tensor.
    Parameters:
        x (Tensor): the input tensor with 2 or more dims.
        n (int): the length of the stripe.
        w (int): the width of the stripe.
        offset (tuple): the offset of the first two dims.
        dim (int): 0 if returns a horizontal stripe; 1 else.
    Example::
    >>> x = torch.arange(25).view(5, 5)
    >>> x
    tensor([[ 0,  1,  2,  3,  4],
            [ 5,  6,  7,  8,  9],
            [10, 11, 12, 13, 14],
            [15, 16, 17, 18, 19],
            [20, 21, 22, 23, 24]])
    >>> stripe(x, 2, 3, (1, 1))
    tensor([[ 6,  7,  8],
            [12, 13, 14]])
    >>> stripe(x, 2, 3, dim=0)
    tensor([[ 0,  5, 10],
            [ 6, 11, 16]])
    '''
    x, seq_len = x.contiguous(), x.size(1)
    stride, numel = list(x.stride()), x[0, 0].numel()
    stride[0] = (seq_len + 1) * numel
    stride[1] = (1 if dim == 1 else seq_len) * numel
    return x.as_strided(size=(n, w, *x.shape[2:]),
                        stride=stride,
                        storage_offset=(offset[0] * seq_len + offset[1]) * numel)



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
from typing import Tuple

import torch
from . import Metric
from ltp.utils import length_to_mask


class TreeMetrics(Metric, alias='tree'):
    def __init__(self, eisner=False):
        """
        Tree Metric(LAS, UAS)

        :param pad_value: 被忽略的目标值
        """
        super(TreeMetrics, self).__init__(LAS=0., UAS=0.)
        self._eisner = eisner
        self._head_true = 0
        self._label_true = 0
        self._union_true = 0
        self._all = 0

    @property
    def UAS(self):
        return (self._head_true / self._all) if self._all != 0 else 0

    @property
    def LAS(self):
        return (self._union_true / self._all) if self._all != 0 else 0

    def step(self, y_pred: Tuple[torch.Tensor, torch.Tensor, torch.Tensor], y: Tuple[torch.Tensor, torch.Tensor]):
        arc_pred, label_pred, seq_len = y_pred
        mask = length_to_mask(seq_len + 1)
        mask[:, 0] = False
        if self._eisner:
            from ltp.utils import eisner
            arc_pred = eisner(arc_pred, mask)
        else:
            arc_pred = torch.argmax(arc_pred, dim=-1)
        label_pred = torch.argmax(label_pred, dim=-1)

        arc_real, label_real = y
        label_pred = label_pred.gather(-1, arc_pred.unsqueeze(-1)).squeeze(-1)

        mask = mask.narrow(-1, 1, mask.size(1) - 1)
        arc_pred = arc_pred.narrow(-1, 1, arc_pred.size(1) - 1)
        label_pred = label_pred.narrow(-1, 1, label_pred.size(1) - 1)

        head_true = (arc_pred == arc_real)[mask]
        label_true = (label_pred == label_real)[mask]

        self._head_true += torch.sum(head_true).item()
        self._label_true += torch.sum(label_true).item()
        self._union_true += torch.sum(label_true[head_true]).item()
        self._all += torch.sum(mask).item()

    def clear(self):
        self._head_true = 0
        self._label_true = 0
        self._union_true = 0
        self._all = 0

    def compute(self):
        return {'LAS': self.LAS, 'UAS': self.UAS}



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
from ltp.core import Registrable


class Callback(metaclass=Registrable):
    """
    回调函数基类
    """
    iteration: int
    epoch: int

    def __init__(self, iteration, epoch):
        self.iteration = iteration
        self.epoch = epoch

    def __call__(self, executor):
        self.call(executor)

    def init(self, executor):
        raise NotImplementedError()

    def call(self, executor):
        raise NotImplementedError()


from .csv_writer import CsvWriter
from .metric import MetricCallback
from .validation import ValidationCallback
from .tensorboard import TensorboardCallback

__all__ = ['Callback', 'CsvWriter', 'TensorboardCallback']



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
import functools
import os
from torchtext.data import Batch


class TaskConfig(object):
    """
    和具体Trainer相关的配置文件
    """

    def __init__(self, config, global_config):
        self.epoch = global_config.epoch
        self.model = config.get("Model")
        self.ckptcfg = config.get("Checkpoint")
        self.fields = config.get("Fields", [])
        self.dataset = config.get("Dataset", None)
        self.loss = config.get("Loss", None)
        self.predictor = config.get("Predictor", None)
        self.metrics = config.get("Metrics", [])
        self.optimizer = config.get("Optimizer")
        self.scheduler = config.get("Scheduler", None)


class Task(object):
    """模型的训练任务
    """

    def __init__(self, config: TaskConfig, device):
        from ltp.models import Model
        from ltp.data import Field
        from ltp.eval import Metric
        from ltp.train import Loss, CheckpointManager, Optimizer

        self.device = device
        self.config = config
        self.model = Model.from_params(config.model)
        self.model.to(self.device)

        self.fields = []

        for field_config in config.fields:
            if not len(field_config):
                continue
            if "name" not in field_config['init']:
                name = field_config["name"]
                field_config.setdefault("name", default=name)
            field = Field.from_params(field_config)
            self.fields.append((field.name, field))

        self.loss = Loss.from_params(config.loss)
        self.metrics = [Metric.from_params(metric, extra={'fields': self.fields})
                        for metric in self.config.metrics if len(metric)]

        self.optimizer = Optimizer.from_params(self.config.optimizer, extra={'model': self.model})

        self.scheduler = None
        self.scheduler_type = None

        self.ckptmgr = CheckpointManager(**config.ckptcfg)

    def build_scheduler(self, num_training_steps):
        from ltp.train import Scheduler
        if self.config.scheduler:
            self.scheduler_type = self.config.scheduler.get('type', 'epoch')
            self.scheduler = Scheduler.from_params(self.config.scheduler, self.optimizer, extra={
                'num_training_steps': num_training_steps
            })
        else:
            self.scheduler_type = None
            self.scheduler = None

    def train_dataset(self, batch_size, shuffle=True):
        from ltp.data import Dataset, DataLoader

        path = self.config.dataset['path']
        train = Dataset.from_params(
            self.config.dataset,
            path=path,
            file=self.config.dataset['train'],
            fields=self.fields
        )

        for name, field in self.fields:
            if not hasattr(field, 'use_vocab') or not field.use_vocab:
                continue
            if not hasattr(field, 'vocab'):
                field.build_vocab(train)

        if 'validation' in self.config.dataset:
            valid = Dataset.from_params(
                self.config.dataset,
                path=path,
                file=self.config.dataset['validation'],
                fields=self.fields
            )
        else:
            ratio = self.config.dataset.get('ratio', 0.9)
            train, valid = train.split(ratio)

        def dataloader(dataset, class_):
            collate_fn = functools.partial(Batch, dataset=dataset, device=self.device)
            dataloader_func = functools.partial(class_, shuffle=shuffle,
                                                batch_size=batch_size,
                                                collate_fn=collate_fn)
            return dataloader_func(dataset)

        train_dataloader = dataloader(train, DataLoader)
        valid_dataloader = dataloader(valid, DataLoader)
        return train_dataloader, valid_dataloader

    def load_dataset(self, file=None, batch_size=10, shuffle=False):
        from ltp.data import Dataset
        from torch.utils.data import DataLoader

        path = file or self.config.dataset['path']
        test = Dataset.from_params(
            self.config.dataset,
            path=path,
            file=self.config.dataset['test'],
            fields=self.fields
        )
        collate_fn = functools.partial(Batch, dataset=test, device=self.device)
        dataloader = functools.partial(DataLoader, batch_size=batch_size, collate_fn=collate_fn, shuffle=shuffle)
        return dataloader(test)

    def restore(self, state=None):
        self.ckptmgr['restore'].load(self, state)

    def load(self, state=None, path=None):
        if path:
            self.ckptmgr[path].load(self, state)
        else:
            self.ckptmgr['best'].load(self, state)

    def save(self, state):
        self.ckptmgr.save(self, state)



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
from typing import Tuple

import torch
from . import Metric
from ltp.utils import length_to_mask


class GraphMetrics(Metric, alias='graph'):
    """
    Graph Parser Metric(f1)
    """

    def __init__(self):
        """
        Graph Parser Metric(f1)
        """
        super(GraphMetrics, self).__init__(f1=0., p=0., r=0.)
        self.nb_correct = 0
        self.nb_pred = 0
        self.nb_true = 0

    def get_entities(self, arcs, labels):
        arcs = torch.nonzero(arcs).cpu().detach().numpy()
        labels = labels.cpu().detach().numpy()

        res = []
        for arc in arcs:
            arc = tuple(arc)
            label = labels[arc]
            res.append((*arc, label))

        return set(res)

    def step(self, y_pred: Tuple[torch.Tensor, torch.Tensor, torch.Tensor], y: Tuple[torch.Tensor, torch.Tensor]):
        arc_pred, label_pred, seq_len = y_pred
        arc_real, label_real = y

        arc_real = arc_real > 0.5
        arc_pred = torch.sigmoid(arc_pred) > 0.5
        # to [B, L+1, L+1]
        label_pred = torch.argmax(label_pred, dim=-1)

        mask = length_to_mask(seq_len + 1)
        mask[:, 0] = False  # ignore the first token of each sentence
        mask = mask.unsqueeze(-1).expand_as(arc_pred)

        arc_pred[mask == False] = False

        true_entities = self.get_entities(arc_real, label_real)
        pred_entities = self.get_entities(arc_pred, label_pred)

        self.nb_correct += len(true_entities & pred_entities)
        self.nb_pred += len(pred_entities)
        self.nb_true += len(true_entities)

    @property
    def precision(self):
        return self.nb_correct / self.nb_pred if self.nb_pred > 0 else 0

    @property
    def recall(self):
        return self.nb_correct / self.nb_true if self.nb_true > 0 else 0

    @property
    def f1beta(self):
        p = self.precision
        r = self.recall
        score = 2 * p * r / (p + r) if (p + r > 0) else 0
        return score

    def compute(self):
        return {'f1': self.f1beta, 'p': self.precision, 'r': self.recall}

    def clear(self):
        self.nb_correct = 0
        self.nb_pred = 0
        self.nb_true = 0



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
import torch
from . import TemperatureScheduler


class ConstantScheduler(TemperatureScheduler, alias='constant'):
    def forward(self, logits_S, logits_T, base_temperature):
        return base_temperature


class FlswScheduler(TemperatureScheduler, alias='flsw'):
    def __init__(self, beta=1, gamma=1, eps=1e-4):
        self.beta = beta
        self.gamma = gamma
        self.eps = eps

    def forward(self, logits_S, logits_T, base_temperature):
        v = logits_S.detach()
        t = logits_T.detach()
        with torch.no_grad():
            v = v / (torch.norm(v, dim=-1, keepdim=True) + self.eps)
            t = t / (torch.norm(t, dim=-1, keepdim=True) + self.eps)
            w = torch.pow((1 - (v * t).sum(dim=-1)), self.gamma)
            tau = base_temperature + (w.mean() - w) * self.beta
        return tau


class CwsmScheduler(TemperatureScheduler, alias='cwsm'):
    def __init__(self, beta=1):
        self.beta = beta

    def forward(self, logits_S, logits_T, base_temperature):
        v = logits_S.detach()
        with torch.no_grad():
            v = torch.softmax(v, dim=-1)
            v_max = v.max(dim=-1)[0]
            w = 1 / (v_max + 1e-3)
            tau = base_temperature + (w.mean() - w) * self.beta
        return tau



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
import torch.nn.init as init
from torch import nn


def initial_parameter(net, initial_method=None):
    """用于初始化 Pytorch 模型内部权重的函数

    Args:
        net: Pytorch Module
        initial_method(str): 取值如下

            - xavier_uniform
            - xavier_normal (default)
            - kaiming_normal, or msra
            - kaiming_uniform
            - orthogonal
            - sparse
            - normal
            - uniform
    """

    if initial_method == 'xavier_uniform':
        init_method = init.xavier_uniform_
    elif initial_method == 'xavier_normal':
        init_method = init.xavier_normal_
    elif initial_method == 'kaiming_normal' or initial_method == 'msra':
        init_method = init.kaiming_normal_
    elif initial_method == 'kaiming_uniform':
        init_method = init.kaiming_uniform_
    elif initial_method == 'orthogonal':
        init_method = init.orthogonal_
    elif initial_method == 'sparse':
        init_method = init.sparse_
    elif initial_method == 'normal':
        init_method = init.normal_
    elif initial_method == 'uniform':
        init_method = init.uniform_
    else:
        init_method = init.xavier_normal_

    def weights_init(m):
        # classname = m.__class__.__name__
        if isinstance(m, nn.Conv2d) or isinstance(m, nn.Conv1d) or isinstance(m, nn.Conv3d):  # for all the cnn
            if initial_method is not None:
                init_method(m.weight.data)
            else:
                init.xavier_normal_(m.weight.data)
            init.normal_(m.bias.data)
        elif isinstance(m, nn.LSTM):
            for w in m.parameters():
                if len(w.data.size()) > 1:
                    init_method(w.data)  # weight
                else:
                    init.normal_(w.data)  # bias
        elif m is not None and hasattr(m, 'weight') and \
                hasattr(m.weight, "requires_grad"):
            init_method(m.weight.data)
        else:
            for w in m.parameters():
                if w.requires_grad:
                    if len(w.data.size()) > 1:
                        init_method(w.data)  # weight
                    else:
                        init.normal_(w.data)  # bias
                # print("init else")

    net.apply(weights_init)





#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

from collections import Counter, OrderedDict
from itertools import chain
from typing import List

import torch

from torchtext.vocab import Vocab
from ltp.const import PAD
from . import Field
from ltp.data.dataset import Dataset


class GraphField(Field, alias='graph'):
    vocab_cls = Vocab

    def __init__(self, name, edge_spliter, pad: str = PAD, tag_spliter=None, use_vocab=True, **kwargs):
        """Graph Field

        Args:
            name(str): field name
            edge_spliter(str): 对边进行拆分的符号
            pad(str): 没有 tag 的地方使用 pad 填充
            tag_spliter(str): 对 tag 进行拆分的符号
            use_vocab(bool): 是否使用词典
            **kwargs:
        """
        super().__init__(name, **kwargs)
        self.edge_spliter = edge_spliter
        self.tag_spliter = tag_spliter
        self.use_vocab = (self.tag_spliter is not None) and use_vocab
        self.pad = pad
        if self.pad:
            self.pad_bias = True
        else:
            self.pad_bias = False

    def preprocess(self, arcs: List[str]):
        node_size = len(arcs)
        res = [arcs_info.split(self.edge_spliter) for arcs_info in arcs]
        if self.use_vocab:
            res = [[(index + 1, *arc.split(self.tag_spliter)) for arc in arcs] for index, arcs in enumerate(res)]
            res = [(int(src), int(target), label) for src, target, label in chain(*res)]
        else:
            res = [[(index + 1, int(arc)) for arc in arcs] for index, arcs in enumerate(res)]
        return node_size + 1, res

    def build_vocab(self, *args, **kwargs):
        counter = Counter()
        sources = []
        for arg in args:
            if isinstance(arg, Dataset):
                sources += [getattr(arg, name) for name, field in
                            arg.fields.items() if field is self]
            else:
                sources.append(arg)
        for data in sources:
            for x in data:
                tags = [arc[-1] for arc in x[1]]
                try:
                    counter.update(tags)
                except TypeError:
                    print("error")
                    counter.update(chain.from_iterable(tags))
        specials = list(
            OrderedDict.fromkeys(
                tok for tok in [self.pad] + kwargs.pop('specials', []) if tok is not None)
        )
        self.vocab = self.vocab_cls(counter, specials=specials, **kwargs)

    def process(self, batch, device=None, *args, **kwargs):
        length, batched_edges = zip(*batch)
        max_edge_size = max(length)

        edge_tensor = torch.zeros(size=(len(length), max_edge_size, max_edge_size))
        tag_tensor = torch.zeros(size=(len(length), max_edge_size, max_edge_size), dtype=torch.long)

        for index, edges in enumerate(batched_edges):
            for edge in edges:
                edge_tensor[index, edge[0], edge[1]] = 1
                tag_tensor[index, edge[0], edge[1]] = self.vocab[edge[2]]

        return edge_tensor.to(device), tag_tensor.to(device) - 1



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
# 参数弃用

def deprecated(info, force=False):
    import sys

    def decorator(func):
        def func_wrapper(*args, **kwargs):
            if force:
                raise NotImplementedError(info)
            print('弃用警告: ', info, file=sys.stderr)
            return func(*args, **kwargs)

        return func_wrapper

    return decorator


class deprecated_param(object):
    def __init__(self, deprecated_args, version, reason):
        self.deprecated_args = set(deprecated_args.split())
        self.version = version
        self.reason = reason

    def __call__(self, callable):
        def wrapper(*args, **kwargs):
            found = self.deprecated_args.intersection(kwargs)
            if found:
                raise TypeError("Parameter(s) %s deprecated since version %s; %s" % (
                    ', '.join(map("'{}'".format, found)), self.version, self.reason))
            return callable(*args, **kwargs)

        return wrapper



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
from . import activation
from .activation import *
from .crf import CRF
from .mlp import MLP
from .ffnn import ffnn
from .bilinear import Bilinear
from .biaffine import Biaffine
from .relative_transformer import RelativeTransformer

from transformers import BertModel, XLNetModel

__all__ = [
    'Swish', 'HSwish', 'Mish', 'MLP', 'ffnn', 'Bilinear', 'Biaffine', 'RelativeTransformer'
]



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

from torch.utils.data import DataLoader


class InfiniteDataLoader(DataLoader):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Initialize an iterator over the dataset.
        self.dataset_iterator = super().__iter__()

    def __iter__(self):
        return self

    def __next__(self):
        try:
            batch = next(self.dataset_iterator)
        except StopIteration:
            # Dataset exhausted, use a new fresh iterator.
            self.dataset_iterator = super().__iter__()
            batch = next(self.dataset_iterator)
        return batch



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
import signal
import fire
from ltp.exe import Config, Executor


class Command(object):
    """
    命令行交互指令
    """
    config: Config
    executor: Executor

    def __init__(self, config, device=None):
        self.config = Config(config, device)
        if self.config.torch_seed is not None:
            self.setup_seed(self.config.torch_seed)
        self.executor = Executor(config=self.config)
        signal.signal(signal.SIGINT, self.__graceful_exit)

    def setup_seed(self, seed: int):
        import torch
        import numpy as np
        import random

        torch.manual_seed(seed)
        np.random.seed(seed)
        random.seed(seed)
        if torch.cuda.is_available():
            torch.manual_seed(seed)
            torch.cuda.manual_seed_all(seed)
            torch.backends.cudnn.benchmark = False
            torch.backends.cudnn.deterministic = True

    def train(self, epoch: int = None):
        """
        进行训练操作

        :param epoch: 训练的轮数，默认10轮或者从配置文件中指定
        """
        self.executor.train(epoch)

    def eval(self, file: str = None, checkpoint: str = None, task: str = None):
        """
        进行 Evaluation 操作

        :param task: 任务，默认为default
        :param checkpoint: 使用的 checkpoint，默认是采用Best ckpt（需要启用Checkpoint Mannger best plugin）
        :param file: 要进行 evaluation 的文件，默认是test数据集
        """
        self.executor.evaluate(file, checkpoint, task=task)

    def predict(self, input: str, output: str, checkpoint: str = None, task: str = 'default'):
        """
        进行预测操作

        :param task: 任务，默认为default
        :param input: 输入文件
        :param output: 输出文件
        :param checkpoint:  使用的 checkpoint，默认是采用Best ckpt（需要启用Checkpoint Mannger best plugin）
        """
        self.executor.predict(input, output, checkpoint, task=task)

    def deploy(self, path: str = 'deploy.model', vocab: str = None):
        """
        将模型中不需要的部分都清理掉，仅仅保留需要预测的部分
        :param path: 最终保存的模型的路径
        :param vocab: 词典名字，从 huggingface 加载
        """
        self.executor.deploy(path, vocab)

    def test(self):
        """
        测试配置文件是否正确，目前只对Dataset进行验证。
        """
        self.executor.test()

    def __graceful_exit(self, signum, frame):
        print("Sig %s caught. Graceful exit has been called. Currently running epoch will be finished." % signum)
        self.executor.stop_condition = lambda state: True


if __name__ == '__main__':
    fire.Fire(Command)



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
from functools import partial
from torch import nn
import ltp.nn


def ffnn(input_size, output_size, hidden_size=None, dropout=None, depth=1, last_dropout=False, **kwargs):
    activation_func = ltp.nn.LeakyReLU
    if hidden_size is None:
        hidden_size = output_size
    for activation, args in kwargs.items():
        activation_func = partial(getattr(ltp.nn.activation, activation), **args)

    module = []
    for i in range(depth - 1):
        module.extend(
            [
                nn.Linear(input_size if i == 0 else hidden_size, hidden_size),
                activation_func()
            ] + ([] if dropout is None else [nn.Dropout(p=dropout, inplace=True)])
        )
    module.append(nn.Linear(hidden_size if len(module) > 0 else input_size, output_size))
    if last_dropout:
        module.append(nn.Dropout(p=dropout, inplace=True))
    return nn.Sequential(*module)



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
import os
import csv
import time

from . import Callback

default_save_directory = './log'
default_filename = 'log.csv'


class CsvWriter(Callback, alias="csv"):
    def __init__(self, iteration, epoch, save_directory=default_save_directory, filename=default_filename,
                 delimiter=',', extra_header=None, data_extractor=lambda x: list()):
        super(CsvWriter, self).__init__(iteration, epoch)
        file, ext = filename.rsplit('.', 1)
        self.log_file_path = os.path.join(save_directory, file + '_' + time.strftime("%Y%m%d_%H%M%S") + '.' + ext)
        self.delimiter = delimiter
        self.data_extractor = data_extractor

        os.makedirs(save_directory, exist_ok=True)

        with open(self.log_file_path, mode='w') as writer:
            self.header = ['timestamp', 'epoch', 'iteration', 'train loss']
            if extra_header is not None:
                if not isinstance(extra_header, list):
                    raise TypeError("extra_header should be a list.")
                self.header += extra_header

            writer = csv.writer(writer, delimiter=delimiter, quotechar='"', quoting=csv.QUOTE_MINIMAL)
            writer.writerow(self.header)

    def call(self, executor):
        self.__save(executor.trainer.state)

    def __save(self, executor_state):
        with open(self.log_file_path, mode='a') as writer:
            writer = csv.writer(writer, delimiter=self.delimiter, quotechar='"', quoting=csv.QUOTE_MINIMAL)
            extra_data = self.data_extractor(executor_state)
            data = [time.time(),
                    executor_state.current_epoch + 1,
                    executor_state.current_iteration + 1,
                    executor_state.last_train_loss] + extra_data
            assert len(data) == len(self.header)
            writer.writerow(data)



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

from typing import Union, Dict
from itertools import chain
import torch, torch.nn.utils.rnn as rnn, numpy as np
from transformers import AutoTokenizer, PreTrainedTokenizer

from ltp.core.exceptions import DataUnsupported
from . import Field


class TextField(Field, alias='text'):
    """文本域

    Args:
        tokenizer: Tokenizer
        name: Field Na,e
        return_length: 是否同时返回 length
        word_info: 是否返回 word index
        is_target: 是否为目标域
    """
    tokenizer_cls = AutoTokenizer

    def __init__(self, tokenizer: Union[str, PreTrainedTokenizer, Dict[str, str]],
                 name='text', return_tokens=False, return_word_idn=False, return_length=True,
                 word_info=True, is_target=False):
        super(TextField, self).__init__(name, is_target)
        self.word_info = word_info
        self.return_length = return_length
        self.return_tokens = return_tokens
        self.return_word_idn = return_word_idn

        if isinstance(tokenizer, str):
            self.tokenizer = self.tokenizer_cls.from_pretrained(tokenizer, use_fast=True)
        elif isinstance(tokenizer, PreTrainedTokenizer):
            self.tokenizer = tokenizer
            assert self.tokenizer.is_fast
        elif isinstance(tokenizer, Dict):
            tokenizer['pretrained_model_name_or_path'] = tokenizer.pop('path')
            tokenizer['use_fast'] = True
            self.tokenizer = self.tokenizer_cls.from_pretrained(**tokenizer)

    def preprocess(self, x):
        sentence = self.tokenizer.batch_encode_plus(
            x, add_special_tokens=False, return_attention_masks=False, return_token_type_ids=False
        )
        subword_len = [len(subword.offsets) for subword in sentence.encodings]
        word_lengths = np.cumsum([0] + subword_len, dtype=np.int64)
        word_start, text_length = word_lengths[:-1], word_lengths[-1]
        word_start_idn = list(chain.from_iterable([0] + [1] * (length - 1) for length in subword_len))

        if text_length > 510:
            raise DataUnsupported("文本过长！！")

        # mixed_sentence = ' '.join(x)
        # resplit = self.tokenizer.encode(
        #     mixed_sentence, add_special_tokens=False, return_attention_masks=False, return_token_type_ids=False
        # )
        # if len(resplit) != len(word_start_idn):
        #     print("X: ", x)
        #     print("Mixed: ", mixed_sentence)

        return ' '.join(x), torch.as_tensor(text_length), \
               torch.as_tensor(word_start_idn), torch.as_tensor(word_start), torch.as_tensor(len(x)),

    def process(self, batch, device=None):
        sentence, text_length, word_start_idn, word_index, word_length = zip(*batch)
        tokenized = self.tokenizer.batch_encode_plus(list(sentence), return_tensors='pt')

        res = {
            'input_ids': tokenized['input_ids'].to(device),
            'token_type_ids': tokenized['token_type_ids'].to(device),
            'attention_mask': tokenized['attention_mask'].to(device),
        }
        if self.return_length:
            res['text_length'] = torch.stack(text_length).to(device)
        if self.return_word_idn:
            res['word_idn'] = rnn.pad_sequence(word_start_idn, batch_first=True).to(device)
        if self.word_info:
            res['word_index'] = rnn.pad_sequence(word_index, batch_first=True).to(device)
            if self.return_length:
                res['word_length'] = torch.stack(word_length).to(device)

        return res


class MixedTextField(TextField):
    def process(self, batch, device=None):
        dataset, sentence, text_length, word_start_idn, word_index, word_length = zip(*batch)
        cls = torch.as_tensor(self.tokenizer.convert_tokens_to_ids(dataset), device=device).unsqueeze_(1)
        tokenized = self.tokenizer.batch_encode_plus(list(sentence), return_tensors='pt')
        input_ids = tokenized['input_ids'].to(device=device)[:, 1:]

        res = {
            'input_ids': torch.cat([cls, input_ids], dim=-1),
            'token_type_ids': tokenized['token_type_ids'].to(device),
            'attention_mask': tokenized['attention_mask'].to(device),
        }
        if self.return_length:
            res['text_length'] = torch.stack(text_length).to(device)
        if self.return_word_idn:
            res['word_idn'] = rnn.pad_sequence(word_start_idn, batch_first=True).to(device)
        if self.word_info:
            res['word_index'] = rnn.pad_sequence(word_index, batch_first=True).to(device)
            if self.return_length:
                res['word_length'] = torch.stack(word_length).to(device)

        return res



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
from . import Optimizer


def BertAdamW4CRF(named_params, lr=1e-5, weight_decay=0.01,
                  bert_lr=1e-5, bert_weight_decay=None,
                  crf_lr=1e-3, crf_weight_decay=None,
                  **kwargs):
    if bert_lr is None:
        bert_lr = lr
    if bert_weight_decay is None:
        bert_weight_decay = weight_decay
    if crf_weight_decay is None:
        crf_weight_decay = weight_decay

    param_optimizer = list(named_params)
    no_decay = ['bias', 'LayerNorm.bias', 'LayerNorm.weight']
    params = {'in_pretrained': {'decay': [], 'no_decay': []}, 'out_pretrained': {'crf': [], 'no_crf': []}}
    for n, p in param_optimizer:
        is_in_pretrained = 'in_pretrained' if 'pretrained' in n else 'out_pretrained'
        is_no_decay = 'no_decay' if any(nd in n for nd in no_decay) else 'decay'
        is_crf = 'crf' if 'crf' in n else 'no_crf'

        if is_in_pretrained == 'in_pretrained':
            params[is_in_pretrained][is_no_decay].append(p)
        else:
            params[is_in_pretrained][is_crf].append(p)

    optimizer_grouped_parameters = [
        {'params': params['in_pretrained']['decay'], 'weight_decay': bert_weight_decay, 'lr': bert_lr},
        {'params': params['in_pretrained']['no_decay'], 'weight_decay': 0.0, 'lr': bert_lr},
        {'params': params['out_pretrained']['crf'], 'weight_decay': crf_weight_decay, 'lr': crf_lr},
        {'params': params['out_pretrained']['no_crf'], 'weight_decay': weight_decay, 'lr': lr},
    ]

    from transformers import optimization
    return optimization.AdamW(optimizer_grouped_parameters, lr=lr, weight_decay=weight_decay, **kwargs)


def Lamb4CRF(named_params, lr=1e-5, weight_decay=0.01,
             bert_lr=1e-5, bert_weight_decay=None,
             crf_lr=1e-3, crf_weight_decay=None,
             **kwargs):
    if bert_lr is None:
        bert_lr = lr
    if bert_weight_decay is None:
        bert_weight_decay = weight_decay
    if crf_weight_decay is None:
        crf_weight_decay = weight_decay

    param_optimizer = list(named_params)
    no_decay = ['bias', 'LayerNorm.bias', 'LayerNorm.weight']
    params = {'in_pretrained': {'decay': [], 'no_decay': []}, 'out_pretrained': {'crf': [], 'no_crf': []}}
    for n, p in param_optimizer:
        is_in_pretrained = 'in_pretrained' if 'pretrained' in n else 'out_pretrained'
        is_no_decay = 'no_decay' if any(nd in n for nd in no_decay) else 'decay'
        is_crf = 'crf' if 'crf' in n else 'no_crf'

        if is_in_pretrained == 'in_pretrained':
            params[is_in_pretrained][is_no_decay].append(p)
        else:
            params[is_in_pretrained][is_crf].append(p)

    optimizer_grouped_parameters = [
        {'params': params['in_pretrained']['decay'], 'weight_decay': bert_weight_decay, 'lr': bert_lr},
        {'params': params['in_pretrained']['no_decay'], 'weight_decay': 0.0, 'lr': bert_lr},
        {'params': params['out_pretrained']['crf'], 'weight_decay': crf_weight_decay, 'lr': crf_lr},
        {'params': params['out_pretrained']['no_crf'], 'weight_decay': weight_decay, 'lr': lr},
    ]

    from torch_optimizer import Lamb
    return Lamb(optimizer_grouped_parameters, lr=lr, weight_decay=weight_decay, **kwargs)


Optimizer.weak_register("BertAdamW4CRF", BertAdamW4CRF)
Optimizer.weak_register("Lamb4CRF", Lamb4CRF)



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
import os
import torch
import itertools
import regex as re
from typing import List

from transformers import AutoTokenizer, cached_path, TensorType, BatchEncoding
from transformers.file_utils import is_remote_url

from ltp.models import Model
from ltp.utils import length_to_mask, eisner, split_sentence
from ltp.utils import USE_PLUGIN, get_entities, is_chinese_char, segment_decode
from ltp.utils import Trie

try:
    from torch.hub import _get_torch_home

    torch_cache_home = _get_torch_home()
except ImportError:
    torch_cache_home = os.path.expanduser(
        os.getenv("TORCH_HOME", os.path.join(os.getenv("XDG_CACHE_HOME", "~/.cache"), "torch"))
    )

default_cache_path = os.path.join(torch_cache_home, "ltp")
LTP_CACHE = os.getenv("LTP_CACHE", default_cache_path)

model_map = {
    'base': 'http://39.96.43.154/ltp/v2/base.tgz',
    'small': 'http://39.96.43.154/ltp/v2/small.tgz',
    'tiny': 'http://39.96.43.154/ltp/v2/tiny.tgz'
}


def no_gard(func):
    def wrapper(*args, **kwargs):
        with torch.no_grad():
            return func(*args, **kwargs)

    return wrapper


WORD_START = 'B-W'
WORD_MIDDLE = 'I-W'


def get_entities_with_list(labels_, itos):
    res = []
    for labels in labels_:
        labels = [itos[label] for label in labels]
        labels = get_entities(labels)
        res.append(labels)
    return res


def get_graph_entities(arcs, labels, itos):
    sequence_num = labels.size(0)
    arcs = torch.nonzero(arcs, as_tuple=False).cpu().detach().numpy().tolist()
    labels = labels.cpu().detach().numpy()

    res = [[] for _ in range(sequence_num)]
    for idx, arc_s, arc_e in arcs:
        label = labels[idx, arc_s, arc_e]
        res[idx].append((arc_s, arc_e, itos[label]))

    return res


def convert_idx_to_name(y, array_len, id2label):
    if id2label:
        return [[id2label[idx] for idx in row[:row_len]] for row, row_len in zip(y, array_len)]
    else:
        return [[idx for idx in row[:row_len]] for row, row_len in zip(y, array_len)]


class LTP(object):
    model: Model
    seg_vocab: List[str]
    pos_vocab: List[str]
    ner_vocab: List[str]
    dep_vocab: List[str]
    sdp_vocab: List[str]
    srl_vocab: List[str]

    tensor: TensorType = TensorType.PYTORCH

    def __init__(self, path: str = 'small', device=None, **kwargs):
        if device is not None:
            if isinstance(device, torch.device):
                self.device = device
            elif isinstance(device, str):
                self.device = torch.device(device)
        elif torch.cuda.is_available():
            self.device = torch.device('cuda')
        else:
            self.device = torch.device('cpu')

        if path in model_map or is_remote_url(path) or os.path.isfile(path):
            proxies = kwargs.pop("proxies", None)
            cache_dir = kwargs.pop("cache_dir", LTP_CACHE)
            force_download = kwargs.pop("force_download", False)
            resume_download = kwargs.pop("resume_download", False)
            local_files_only = kwargs.pop("local_files_only", False)
            path = cached_path(
                model_map.get(path, path),
                cache_dir=cache_dir,
                force_download=force_download,
                proxies=proxies,
                resume_download=resume_download,
                local_files_only=local_files_only,
                extract_compressed_file=True
            )
        elif not os.path.isdir(path):
            raise FileNotFoundError()
        ckpt = torch.load(os.path.join(path, "ltp.model"), map_location=self.device)
        ckpt['model_config']['init'].pop('pretrained')
        self.cache_dir = path
        self.model = Model.from_params(ckpt['model_config'], config=ckpt['pretrained_config']).to(self.device)
        self.model.load_state_dict(ckpt['model'])
        self.model.eval()
        # todo fp16
        self.seg_vocab = [WORD_START, WORD_MIDDLE]
        self.pos_vocab = ckpt['pos']
        self.ner_vocab = ckpt['ner']
        self.dep_vocab = ckpt['dep']
        self.sdp_vocab = ckpt['sdp']
        self.srl_vocab = [re.sub(r'ARG(\d)', r'A\1', tag.lstrip('ARGM-')) for tag in ckpt['srl']]
        self.tokenizer = AutoTokenizer.from_pretrained(path, config=self.model.pretrained.config, use_fast=True)
        self.trie = Trie()

    def __str__(self):
        return f"LTP {self.version} on {self.device}"

    def __repr__(self):
        return f"LTP {self.version} on {self.device}"

    @property
    def version(self):
        from ltp import __version__ as version
        return version

    def init_dict(self, path, max_window=None):
        self.trie.init(path, max_window)

    def add_words(self, words, max_window=4):
        self.trie.add_words(words)
        self.trie.max_window = max_window

    @staticmethod
    def sent_split(inputs: List[str], flag: str = "all", limit: int = 510):
        inputs = [split_sentence(text, flag=flag, limit=limit) for text in inputs]
        inputs = list(itertools.chain(*inputs))
        return inputs

    def seg_with_dict(self, inputs: List[str], tokenized: BatchEncoding):
        # 进行正向字典匹配
        matching = []
        for source_text, encoding in zip(inputs, tokenized.encodings):
            text = [source_text[start:end] for start, end in encoding.offsets[1:-1] if end != 0]
            matching_pos = self.trie.maximum_forward_matching(text)
            matching.append(matching_pos)
        return matching

    @no_gard
    def _seg(self, tokenizerd):
        input_ids = tokenizerd['input_ids'].to(self.device)
        attention_mask = tokenizerd['attention_mask'].to(self.device)
        token_type_ids = tokenizerd['token_type_ids'].to(self.device)
        length = torch.sum(attention_mask, dim=-1) - 2

        pretrained_output, *_ = self.model.pretrained(
            input_ids=input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids
        )

        # remove [CLS] [SEP]
        word_cls = pretrained_output[:, :1]
        char_input = torch.narrow(pretrained_output, 1, 1, pretrained_output.size(1) - 2)
        segment_output = torch.argmax(self.model.seg_decoder(char_input), dim=-1).cpu().numpy()
        return word_cls, char_input, segment_output, length

    @no_gard
    def seg(self, inputs: List[str]):
        tokenizerd = self.tokenizer.batch_encode_plus(inputs, return_tensors=self.tensor, padding=True)
        cls, hidden, seg, length = self._seg(tokenizerd)

        # merge segments with maximum forward matching
        if self.trie.is_init:
            matches = self.seg_with_dict(inputs, tokenizerd)
            for sent_match, sent_seg in zip(matches, seg):
                for start, end in sent_match:
                    sent_seg[start] = 0
                    sent_seg[start + 1:end] = 1
                    if end < len(sent_seg):
                        sent_seg[end] = 0

        segment_output = convert_idx_to_name(seg, length, self.seg_vocab)
        if USE_PLUGIN:
            offsets = [list(filter(lambda x: x != (0, 0), encodings.offsets)) for encodings in tokenizerd.encodings]
            words = [list(filter(lambda x: x is not None, encodings.words)) for encodings in tokenizerd.encodings]
            sentences, word_idx, word_length = segment_decode(inputs, segment_output, offsets, words)
            word_idx = [torch.as_tensor(idx, device=self.device) for idx in word_idx]
        else:
            sentences = []
            word_idx = []
            word_length = []

            for source_text, encoding, sentence_seg_tag in zip(inputs, tokenizerd.encodings, segment_output):
                text = [source_text[start:end] for start, end in encoding.offsets[1:-1] if end != 0]

                last_word = 0
                for idx, word in enumerate(encoding.words[1:-1]):
                    if word is None or is_chinese_char(text[idx][-1]):
                        continue
                    if word != last_word:
                        text[idx] = ' ' + text[idx]
                        last_word = word
                    else:
                        sentence_seg_tag[idx] = WORD_MIDDLE

                entities = get_entities(sentence_seg_tag)
                word_length.append(len(entities))
                sentences.append([''.join(text[entity[1]:entity[2] + 1]).strip() for entity in entities])
                word_idx.append(torch.as_tensor([entity[1] for entity in entities], device=self.device))

        word_idx = torch.nn.utils.rnn.pad_sequence(word_idx, batch_first=True)
        word_idx = word_idx.unsqueeze(-1).expand(-1, -1, hidden.shape[-1])  # 展开

        word_input = torch.gather(hidden, dim=1, index=word_idx)  # 每个word第一个char的向量

        word_cls_input = torch.cat([cls, word_input], dim=1)
        word_cls_mask = length_to_mask(torch.as_tensor(word_length, device=self.device) + 1)
        word_cls_mask[:, 0] = False  # ignore the first token of each sentence
        return sentences, {
            'word_cls': cls, 'word_input': word_input, 'word_length': word_length,
            'word_cls_input': word_cls_input, 'word_cls_mask': word_cls_mask
        }

    @no_gard
    def pos(self, hidden: dict):
        # 词性标注
        postagger_output = self.model.pos_decoder(hidden['word_input'], hidden['word_length'])
        postagger_output = torch.argmax(postagger_output, dim=-1).cpu().numpy()
        postagger_output = convert_idx_to_name(postagger_output, hidden['word_length'], self.pos_vocab)
        return postagger_output

    @no_gard
    def ner(self, hidden: dict):
        # 命名实体识别
        word_length = torch.as_tensor(hidden['word_length'], device=self.device)
        ner_output = self.model.ner_decoder(hidden['word_input'], word_length)
        ner_output = torch.argmax(ner_output, dim=-1).cpu().numpy()
        ner_output = convert_idx_to_name(ner_output, hidden['word_length'], self.ner_vocab)
        return [get_entities(ner) for ner in ner_output]

    @no_gard
    def srl(self, hidden: dict, keep_empty=True):
        # 语义角色标注
        word_length = torch.as_tensor(hidden['word_length'], device=hidden['word_input'].device)
        word_mask = length_to_mask(word_length)
        srl_output, srl_length, crf = self.model.srl_decoder(hidden['word_input'], hidden['word_length'])
        mask = word_mask.unsqueeze_(-1).expand(-1, -1, word_mask.size(1))
        mask = (mask & mask.transpose(-1, -2)).flatten(end_dim=1)
        index = mask[:, 0]
        mask = mask[index]

        srl_input = srl_output.flatten(end_dim=1)[index]
        srl_entities = crf.decode(torch.log_softmax(srl_input, dim=-1), mask)
        srl_entities = get_entities_with_list(srl_entities, self.srl_vocab)

        srl_labels_res = []
        for length in srl_length:
            srl_labels_res.append([])
            curr_srl_labels, srl_entities = srl_entities[:length], srl_entities[length:]
            srl_labels_res[-1].extend(curr_srl_labels)

        if not keep_empty:
            srl_labels_res = [[(idx, labels) for idx, labels in enumerate(srl_labels) if len(labels)]
                              for srl_labels in srl_labels_res]
        return srl_labels_res

    @no_gard
    def dep(self, hidden: dict, fast=False):
        # 依存句法树
        dep_arc, dep_label, word_length = self.model.dep_decoder(hidden['word_cls_input'], hidden['word_length'])
        if fast:
            dep_arc_fix = dep_arc.argmax(dim=-1).unsqueeze_(-1).expand_as(dep_arc)
        else:
            dep_arc_fix = eisner(dep_arc, hidden['word_cls_mask']).unsqueeze_(-1).expand_as(dep_arc)
        dep_arc = torch.zeros_like(dep_arc, dtype=torch.bool).scatter_(dim=-1, index=dep_arc_fix, value=True)
        dep_label = torch.argmax(dep_label, dim=-1)

        word_cls_mask = hidden['word_cls_mask']
        word_cls_mask = word_cls_mask.unsqueeze(-1).expand(-1, -1, word_cls_mask.size(1))
        dep_arc = dep_arc & word_cls_mask
        dep_label = get_graph_entities(dep_arc, dep_label, self.dep_vocab)

        return dep_label

    @no_gard
    def sdp(self, hidden: dict, graph=True):
        # 语义依存
        sdp_arc, sdp_label, _ = self.model.sdp_decoder(hidden['word_cls_input'], hidden['word_length'])
        sdp_arc = torch.sigmoid_(sdp_arc)

        if graph:
            # 语义依存图
            sdp_arc.transpose_(-1, -2)
            sdp_root_mask = sdp_arc[:, 0].argmax(dim=-1).unsqueeze_(-1).expand_as(sdp_arc[:, 0])
            sdp_arc[:, 0] = 0
            sdp_arc[:, 0].scatter_(dim=-1, index=sdp_root_mask, value=1)
            sdp_arc_T = sdp_arc.transpose(-1, -2)
            sdp_arc_fix = sdp_arc_T.argmax(dim=-1).unsqueeze_(-1).expand_as(sdp_arc)
            sdp_arc = ((sdp_arc_T > 0.5) & (sdp_arc_T > sdp_arc)). \
                scatter_(dim=-1, index=sdp_arc_fix, value=True)
        else:
            # 语义依存树
            sdp_arc_fix = eisner(sdp_arc, hidden['word_cls_mask']).unsqueeze_(-1).expand_as(sdp_arc)
            sdp_arc = torch.zeros_like(sdp_arc, dtype=torch.bool).scatter_(dim=-1, index=sdp_arc_fix, value=True)

        sdp_label = torch.argmax(sdp_label, dim=-1)

        word_cls_mask = hidden['word_cls_mask']
        word_cls_mask = word_cls_mask.unsqueeze(-1).expand(-1, -1, word_cls_mask.size(1))
        sdp_arc = sdp_arc & word_cls_mask
        sdp_label = get_graph_entities(sdp_arc, sdp_label, self.sdp_vocab)

        return sdp_label



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
from .early_stopping import EarlyStopping


class NoStopping(object):
    """
    永不停止
    """
    def __call__(self, state):
        return False


__all__ = ['NoStopping', 'EarlyStopping']



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
import torch
from torch import nn, Tensor
from ltp.nn import Bilinear, ffnn


class Biaffine(nn.Module):
    def __init__(self, in1_features, in2_features, out_features, bias_x=True, bias_y=True, **kwargs):
        super(Biaffine, self).__init__()
        self.bw = nn.Parameter(torch.as_tensor([0.5, 0.5]), requires_grad=True)
        self.bilinear = Bilinear(in1_features, in2_features, out_features, bias_x=bias_x, bias_y=bias_y, expand=True)
        self.linear = ffnn(in1_features + in2_features, output_size=out_features, **kwargs.pop("linear", {}))

    def forward(self, x1: Tensor, x2: Tensor):
        len1, len2 = x1.size(1), x2.size(1)
        bw_norm = torch.softmax(self.bw, dim=-1)
        bilinear_logits = self.bilinear(x1, x2).permute(0, 2, 3, 1)
        x1 = x1.unsqueeze(2).contiguous().expand([-1, -1, len2, -1])
        x2 = x2.unsqueeze(1).contiguous().expand([-1, len1, -1, -1])
        linear_logits = self.linear(torch.cat([x1, x2], dim=-1))
        return bw_norm[0] * bilinear_logits + bw_norm[1] * linear_logits

    def extra_repr(self):
        return 'in1_features={}, in2_features={}, out_features={}, bias_x={}, bias_y={}'.format(
            self.in1_features, self.in2_features, self.out_features, self.bias_x, self.bias_y
        )



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
from typing import Union

from torch.nn import Module, Linear
from transformers import PretrainedConfig, PreTrainedModel, AutoConfig, AutoModel
from ltp.core import Registrable
from ltp import modules

"""
说明：
    text: 指代类似生文本，指的的文本，也就是字(Char Based)级别
    word: 指代词语，通常是词语，比如 word_index 就是词语开头的下标
"""


class Model(Module, metaclass=Registrable):
    """
    模型基础类，从本类继承的模型自动注册
    """
    pretrained: PreTrainedModel

    def create_pretrained(self, pretrained: str = None, config: Union[str, dict, PretrainedConfig] = None,
                          freeze: bool = False):
        if isinstance(pretrained, str):  # 认为是 PATH 或 Huggingface 社区模型
            pretrained = AutoModel.from_pretrained(pretrained)
        elif isinstance(config, PretrainedConfig):
            pretrained = AutoModel.from_config(config)
        elif isinstance(config, str) or isinstance(config, dict):  # 认为是 PATH 或 Huggingface 社区模型
            config = AutoConfig.for_model(**config)
            pretrained = AutoModel.from_config(config)
        else:
            raise NotImplementedError()

        if freeze:
            for param in pretrained.parameters():
                param.requires_grad = False
        return pretrained

    def create_decoder(self, input_size, label_num, dropout=0.1, **kwargs):
        """
        封装了各种解码器
        """
        decoder_type = kwargs.pop('decoder', 'Linear')
        if decoder_type:
            decoder_args = kwargs.pop(decoder_type, {})
            decoder_args.setdefault('dropout', dropout)
            decoder_class = modules.Module.by_name(decoder_type)
            return decoder_class(input_size, label_num, **decoder_args)
        else:
            raise NotImplementedError()

    def __getitem__(self, item):
        self.task = item
        return self

    def forward(self, *args, **kwargs):
        """前向传播函数
        """
        raise NotImplementedError()



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
import os

from tqdm import tqdm

import re
from . import Dataset
from ltp.data.example import Example
from ltp.data.utils import iter_lines


class LineDataset(Dataset, alias="line"):
    """
    Dataset 注册名: line
    每一行一个句子，词语之间按空格分割，下划线分割词语与词性，例子如下::

        上海_NR	浦东_NR	开发_NN	与_CC	法制_NN	建设_NN	同步_VV

    或者使用其他字符分割(由split参数控制)

        上海/NR	浦东/NR	开发/NN	与/CC	法制/NN	建设/NN	同步/VV

    """

    def __init__(self, path, file, fields, split='\0', maxsplit=0, silent=True, proxy_property=None,
                 line_split=None, line_strip=None, **kwargs):
        filename = os.path.join(path, file)
        examples = []
        split_regex = re.compile(split)
        fields_slices = [idx for idx, field in enumerate(fields) if field is not None]
        if proxy_property is not None:
            field_map = {field[0]: idx for idx, field in enumerate(fields)}
            for proxy, source in proxy_property.items():
                fields_slices[field_map[proxy]] = fields_slices[field_map[source]]

        for line_num, line in tqdm(list(iter_lines(filename, line_split, line_strip))):
            if len(line) == 0:
                continue
            data = [list(item) for item in zip(*(split_regex.split(item, maxsplit) for item in line))]
            data = [data[field_slice] for field_slice in fields_slices]
            try:
                examples.append(Example.fromlist(data, fields))
            except Exception as e:
                if not silent:
                    print(line_num, e, line)

        super(LineDataset, self).__init__(examples, fields, **kwargs)


from ltp.utils import deprecated


@deprecated(info='CTB数据集已经被line代替')
def CTBDataset(path, fields, silent=True, proxy_property=None, split=None, strip=None, **kwargs):
    return LineDataset(path=path, fields=fields, split='\0', silent=silent, proxy_property=proxy_property,
                       line_split=split, line_strip=strip, **kwargs)


Dataset.weak_register('CTB', CTBDataset)



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

from .vocab import Vocab
from .processing import PreProcessing, PostProcessing
from .fields import Field
from .example import Example
from .dataset import Dataset
from .dataloader import DataLoader, InfiniteDataLoader



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: jeffrey
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

import os
from typing import List

import pygtrie


class Trie(pygtrie.CharTrie):
    __is_init: bool
    __max_window: int
    __min_start: int

    def __init__(self, max_window=4, init_path: str = None):
        super(Trie, self).__init__()
        self.__is_init = False
        self.__min_start = max_window
        self.__max_window = max_window
        self.init(init_path)

    @property
    def is_init(self):
        return self.__is_init

    @is_init.setter
    def is_init(self, value: bool):
        if value:
            self.__is_init = True

    @property
    def max_window(self):
        return self.__max_window

    @max_window.setter
    def max_window(self, value: int):
        if isinstance(value, int) and value > 0:
            self.__max_window = value
            self.__min_start = min(self.__min_start, self.__max_window)

    @property
    def min_start(self):
        return self.__min_start

    @min_start.setter
    def min_start(self, value: int):
        if isinstance(value, int) and value > 0:
            self.__min_start = min(value, self.__min_start, self.__max_window)

    def init(self, init_path: str = None, max_window=None):
        self.max_window = max_window
        if init_path and os.path.exists(init_path):
            self.is_init = True
            with open(init_path, "r", encoding="utf-8") as f:
                for word in f.readlines():
                    word = word.strip()
                    if word:
                        self[word] = True

    def add_words(self, words):
        self.is_init = True
        if isinstance(words, str):
            self[words] = True
            self.min_start = len(words)
        if isinstance(words, list):
            for word in words:
                self[word] = True
                self.min_start = len(word)

    def maximum_forward_matching(self, text: List[str]):
        maximum_matching_pos = []
        text_len = len(text)
        for start in range(text_len - 1):
            candidate = None
            for end in range(start + self.min_start, min(text_len, start + self.max_window + 1)):
                if self.get("".join(text[start:end]), False):
                    candidate = (start, end)
            if candidate:
                maximum_matching_pos.append(candidate)
        return maximum_matching_pos



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

from typing import Dict

from ...core import Registrable


class Metric(metaclass=Registrable):
    """
    Metric 基础类，在配置文件中的配置项为::

        [[Metrics]]
        class = "Accuracy"
        [Metrics.init]
        pad_value = -1
    """
    eps = 1e-6
    defaults: Dict[str, float]

    def __init__(self, **defaults):
        """

        """
        self.defaults = defaults

    def step(self, y_pred, y):
        """
        步进一次
        :param y_pred: 预测值
        :param y: 真实值
        """
        raise NotImplementedError()

    def compute(self) -> Dict[str, float]:
        """
        返回 metrics 结果
        :return: Dict[str, float]
        """
        raise NotImplementedError()

    def clear(self):
        """
        把已经统计的Metrics结果清0
        """
        raise NotImplementedError()


from .accuracy import Accuracy
from .common_loss import CommonLoss
from .tree import TreeMetrics
from .graph import GraphMetrics
from .sequence import Sequence
from .biaffine_crf_span import BiaffineCRFSpan
from .segment import Segment

__all__ = [
    'Metric', 'Accuracy', 'CommonLoss', 'TreeMetrics', 'GraphMetrics',
    'Sequence', 'BiaffineCRFSpan'
]



#! /usr/bin/env python
# -*- coding: utf-8 -*_

from distutils.core import setup
import setuptools

from ltp import __version__ as version

with open("README.md", "r", encoding='utf-8') as fh:
    long_description = fh.read()

setup(
    name='ltp',
    version=version,
    author='Yunlong Feng',
    author_email='ylfeng@ir.hit.edu.cn',
    url='https://github.com/HIT-SCIR/ltp',
    description='Language Technology Platform',
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(exclude=['tests', 'examples', 'docs']),
    install_requires=[
        'torch>=1.2.0',
        'torchtext==0.5.0',
        'transformers>=3.0',
        'cytoolz~=0.10.1',
        'pygtrie==2.3.3'
        'tqdm',
        'toml',
        'fire',
    ],
    classifiers=[
        'Development Status :: 1 - Planning',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Software Development :: Libraries'
    ],
    python_requires='>=3.6.*, <4',
)



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
import torch
from . import Metric
import numpy as np
from ltp.utils.seqeval import get_entities


class Sequence(Metric):
    """
    用于命名实体识别或其他Span序列任务
    """

    def __init__(self, id2label, pad_value=-1, suffix=False, no_suffix=False):
        super(Sequence, self).__init__(f1=0., p=0., r=0.)
        self.field = id2label
        self.pad_value = pad_value
        self.suffix = suffix
        self.no_suffix = no_suffix

        self.nb_correct = 0
        self.nb_pred = 0
        self.nb_true = 0
        self._id2label = id2label if isinstance(id2label, list) else None

    @property
    def id2label(self):
        if not self._id2label:
            # Lazy Import
            self._id2label = self.field.vocab.itos[1:]
            if self.no_suffix:
                self._id2label = [tag + '-W' for tag in self.field.vocab.itos[1:]]
            else:
                self._id2label = self.field.vocab.itos[1:]
        return self._id2label

    def convert_idx_to_name(self, y, array_indexes):
        y = [[self.id2label[idx] for idx in row[row_indexes]] for
             row, row_indexes in zip(y, array_indexes)]
        return y

    def predict(self, y_pred: torch.Tensor, y_true: torch.Tensor):
        y_pred = y_pred.cpu().detach().numpy()
        y_true = y_true.cpu().detach().numpy()

        non_pad_indexes = [np.nonzero(y_true_row != self.pad_value)[0] for y_true_row in y_true]

        y_true = self.convert_idx_to_name(y_true, non_pad_indexes)
        y_pred = self.convert_idx_to_name(y_pred, non_pad_indexes)

        return y_pred, y_true

    def step(self, y_pred: torch.Tensor, y: torch.Tensor):
        y_pred = torch.argmax(y_pred, dim=-1)
        y_pred, y = self.predict(y_pred, y)

        true_entities = set(get_entities(y, self.suffix))
        pred_entities = set(get_entities(y_pred, self.suffix))

        self.nb_correct += len(true_entities & pred_entities)
        self.nb_pred += len(pred_entities)
        self.nb_true += len(true_entities)

    @property
    def precision(self):
        return self.nb_correct / self.nb_pred if self.nb_pred > 0 else 0

    @property
    def recall(self):
        return self.nb_correct / self.nb_true if self.nb_true > 0 else 0

    @property
    def f1beta(self):
        p = self.precision
        r = self.recall
        score = 2 * p * r / (p + r) if (p + r > 0) else 0
        return score

    def compute(self):
        return {'f1': self.f1beta, 'p': self.precision, 'r': self.recall}

    def clear(self):
        self.nb_correct = 0
        self.nb_pred = 0
        self.nb_true = 0

    @classmethod
    def from_extra(cls, extra: dict):
        init = extra["config"]['init']
        id2label = init['id2label']

        for field_name, field in extra['fields']:
            if field_name == id2label:
                return {'id2label': field}

        return {'id2label': None}



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
from . import Loss
from .kd_loss import KdMseLoss, KdCeLoss
import torch
import torch.nn.functional as F
from ltp.utils import length_to_mask, select_logits_with_mask


class SegLoss(Loss, alias='seg'):
    def __init__(self, size_average=None, reduce=None, reduction='mean'):
        super(SegLoss, self).__init__(size_average, reduce, reduction)

    def forward(self, inputs, targets):
        mask = length_to_mask(targets['text_length'])
        target = targets['word_idn']
        loss = F.cross_entropy(inputs[mask], target[mask], reduction=self.reduction)
        return loss

    def distill(self, inputs, targets, temperature_calc, distill_loss, gold=None):
        mask = length_to_mask(gold['text_length'])
        logits = inputs[mask]
        logits_T = targets[mask]
        temperature = temperature_calc(logits, logits_T)
        return distill_loss(logits, logits_T, temperature)


class DepLoss(Loss, alias='dep'):
    def __init__(self, size_average=None, ignore_index=-1, loss_interpolation=0.4,
                 reduce=None, reduction='mean'):
        super(DepLoss, self).__init__(size_average, reduce, reduction)
        self.ignore_index = ignore_index
        self.loss_interpolation = loss_interpolation
        self.kd_ce_loss = KdCeLoss()

    def forward(self, inputs, targets):
        arcs, rels = targets
        arc_scores, rel_scores, seq_lens = inputs
        mask = length_to_mask(seq_lens + 1)
        mask[:, 0] = False  # ignore the first token of each sentence

        arc_scores, rel_scores = arc_scores[mask], rel_scores[mask]

        # for taget not bos
        mask = torch.narrow(mask, dim=-1, start=1, length=mask.size(1) - 1)
        arcs, rels = arcs[mask], rels[mask]
        rel_scores = rel_scores[torch.arange(len(arcs)), arcs]

        arc_loss = F.cross_entropy(
            arc_scores, arcs, weight=None,
            ignore_index=self.ignore_index, reduction=self.reduction
        )
        rel_loss = F.cross_entropy(
            rel_scores, rels, weight=None,
            ignore_index=self.ignore_index, reduction=self.reduction
        )
        loss = 2 * ((1 - self.loss_interpolation) * arc_loss + self.loss_interpolation * rel_loss)
        return loss

    def distill(self, inputs, targets, temperature_calc, distill_loss, gold=None):
        arc_scores, rel_scores, seq_lens = inputs
        arc_scores_T, rel_scores_T, _ = targets

        mask = length_to_mask(seq_lens + 1)
        mask[:, 0] = False  # ignore the first token of each sentence

        arc_logits = select_logits_with_mask(arc_scores, mask)
        arc_logits_T = select_logits_with_mask(arc_scores_T, mask)
        arc_temperature = temperature_calc(arc_logits, arc_logits_T)

        rel_logits = select_logits_with_mask(rel_scores, mask)
        rel_logits_T = select_logits_with_mask(rel_scores_T, mask)
        rel_temperature = temperature_calc(rel_logits, rel_logits_T)

        loss = 2 * ((1 - self.loss_interpolation) * self.kd_ce_loss(arc_logits, arc_logits_T, arc_temperature)
                    + self.loss_interpolation * self.kd_ce_loss(rel_logits, rel_logits_T, rel_temperature))

        return loss


class SDPLoss(Loss, alias='sdp'):
    def __init__(self, weight=None, size_average=None, ignore_index=-1, loss_interpolation=0.4,
                 reduce=None, reduction='mean'):
        super(SDPLoss, self).__init__(size_average, reduce, reduction)
        self.weight = weight
        self.ignore_index = ignore_index
        self.loss_interpolation = loss_interpolation
        self.kd_mse_loss = KdMseLoss()
        self.kd_ce_loss = KdCeLoss()

    def forward(self, inputs, targets):
        arcs, rels = targets
        arc_scores, rel_scores, seq_lens = inputs
        mask = length_to_mask(seq_lens + 1, dtype=torch.float)
        mask[:, 0] = 0  # ignore the first token of each sentence

        mask = mask.unsqueeze(-1)
        mask = mask.expand_as(arcs)

        arc_loss = F.binary_cross_entropy_with_logits(
            arc_scores, arcs, weight=mask, reduction=self.reduction
        )

        num_tags = rel_scores.shape[-1]
        rel_loss = F.cross_entropy(
            rel_scores.contiguous().view((-1, num_tags)), rels.contiguous().view(-1), weight=self.weight,
            ignore_index=self.ignore_index, reduction=self.reduction
        )
        loss = 2 * ((1 - self.loss_interpolation) * arc_loss + self.loss_interpolation * rel_loss)
        return loss

    def distill(self, inputs, targets, temperature_calc, distill_loss, gold=None):
        arc_scores, rel_scores, seq_lens = inputs
        arc_scores_T, rel_scores_T, _ = targets

        mask = length_to_mask(seq_lens + 1)
        mask[:, 0] = False

        arc_mask = mask.unsqueeze(-1).expand_as(arc_scores)
        arc_logits = torch.sigmoid(arc_scores)[arc_mask]
        arc_logits_T = torch.sigmoid(arc_scores_T)[arc_mask]
        arc_temperature = temperature_calc(arc_logits, arc_logits_T)

        rel_logits = select_logits_with_mask(rel_scores, mask)
        rel_logits_T = select_logits_with_mask(rel_scores_T, mask)
        rel_temperature = temperature_calc(rel_logits, rel_logits_T)

        loss = 2 * ((1 - self.loss_interpolation) * F.mse_loss(arc_logits / arc_temperature,
                                                               arc_logits_T / arc_temperature)
                    + self.loss_interpolation * self.kd_ce_loss(rel_logits, rel_logits_T, rel_temperature))
        return loss


class BiaffineCRF(Loss, alias='biaffine_crf'):
    def __init__(self, reduction='sum', cross_entropy=False):
        super(BiaffineCRF, self).__init__()
        self.reduction = reduction
        self.cross_entropy = cross_entropy
        self.kd_mse_loss = KdMseLoss()

    def forward(self, inputs, targets):
        emissions, seq_lens, crf = inputs
        rel_gold, rel_gold_set = targets

        mask = length_to_mask(seq_lens)
        mask = mask.unsqueeze_(-1).expand_as(rel_gold)
        mask = mask & mask.transpose(-1, -2)
        mask = mask.flatten(end_dim=1)
        index = mask[:, 0]

        mask = mask[index]
        emissions = emissions.flatten(end_dim=1)[index]
        emissions = F.log_softmax(emissions, dim=-1)
        rel_gold = rel_gold.flatten(end_dim=1)[index]

        if self.cross_entropy:
            cross_entropy = F.cross_entropy(emissions[mask], rel_gold[mask])
            crf_loss = crf.forward(emissions=emissions, tags=rel_gold, mask=mask, reduction=self.reduction)
            return cross_entropy - crf_loss
        else:
            return -crf.forward(emissions=emissions, tags=rel_gold, mask=mask, reduction=self.reduction)

    def distill(self, inputs, targets, temperature_calc, distill_loss, gold=None):
        emissions, seq_lens, crf = inputs
        emissions_T, _, crf_T = targets

        mask = length_to_mask(seq_lens)
        mask = mask.unsqueeze_(-1).expand(-1, -1, mask.size(1))
        mask = mask & mask.transpose(-1, -2)

        logits_loss = F.mse_loss(emissions[mask], emissions_T[mask])
        crf_loss = F.mse_loss(crf.transitions, crf_T.transitions) + \
                   F.mse_loss(crf.start_transitions, crf_T.start_transitions) + \
                   F.mse_loss(crf.end_transitions, crf_T.end_transitions)

        return logits_loss + crf_loss



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
"""
预测器，在配置文件中需配置此项::

    [Predictor]
    class = "simple"
"""

from typing import Dict, Any
from ltp.core import Registrable


class Predictor(metaclass=Registrable):
    """
    若需要使用Predict功能，则需要配置此项::

        [Predictor]
        class = "simple"
            [Predictor.init]
            len_field = 'text_len'

    """
    fields: Dict[str, Any]

    def __init__(self, fields):
        self.fields = dict(fields)

    def __call__(self, *args, **kwargs):
        return self.predict(*args, **kwargs)

    def predict(self, inputs, pred):
        """
        对结果进行预测
        """
        raise NotImplementedError("Predictor 需要实现predict函数")

    @property
    def input_fields(self):
        return [field for field in self.fields.values() if not field.is_target]

    @property
    def target_fields(self):
        return [field for field in self.fields.values() if field.is_target]


from .simple import Simple
from .segment import Segment
from .biaffine import Biaffine

__all__ = ['Simple', 'Segment', 'Biaffine']



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

from .dataset import rationed_split, RandomShuffler, Dataset
from .corpus import CorpusDataset
from .line import LineDataset
from .mixed import MixedDataset

__all__ = ['Dataset', 'LineDataset', 'CorpusDataset', 'MixedDataset']



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>

from collections import Counter, OrderedDict
from itertools import chain
from typing import List

import torch

from torchtext.vocab import Vocab
from ltp.data.dataset import Dataset

from . import Field


class BiaffineField(Field, alias='biaffine'):
    """
    Biaffine 域

    Args:
        name: label name
        use_vocab: 是否使用词典
        pad: 无 label 的 label 位置使用 pad 填充
        labels: 可以给定 label 而不是通过统计数据得到
    """
    vocab_cls = Vocab

    def __init__(self, name, use_vocab=True, pad='O', labels=None, **kwargs):
        super().__init__(name, **kwargs)
        self.pad = pad
        self.use_vocab = use_vocab
        if labels:
            counter = Counter()
            counter.update(labels)
            specials = list(
                OrderedDict.fromkeys(
                    tok for tok in [self.pad] + kwargs.pop('specials', []) if tok is not None)
            )
            self.vocab = self.vocab_cls(counter, specials=specials)

    def preprocess(self, inputs: List[List[str]]):
        heads, *labels = inputs
        sentences_length = len(heads)
        heads = [i for i, head in enumerate(heads) if head == 'Y']

        target_labels = []
        for predicate, label in zip(heads, labels):
            label = [(predicate, index, label) for index, label in enumerate(label)]
            target_labels.extend(label)

        return sentences_length, heads, target_labels

    def build_vocab(self, *args, **kwargs):
        counter = Counter()
        sources = []
        for arg in args:
            if isinstance(arg, Dataset):
                sources += [getattr(arg, name) for name, field in
                            arg.fields.items() if field is self]
            else:
                sources.append(arg)
        for data in sources:
            for x in data:
                tags = [label[-1] for label in x[-1]]
                try:
                    counter.update(tags)
                except TypeError:
                    print("error")
                    counter.update(chain.from_iterable(tags))
        specials = list(
            OrderedDict.fromkeys(
                tok for tok in [self.pad] + kwargs.pop('specials', []) if tok is not None)
        )
        self.vocab = self.vocab_cls(counter, specials=specials, **kwargs)

    def process(self, batch, device=None, *args, **kwargs):
        sentence_lengths, heads, labels = zip(*batch)
        label_set = set()

        batch_size = len(batch)
        max_sentences_length = max(sentence_lengths)

        label_tensor = torch.zeros([batch_size] + [max_sentences_length] * 2, dtype=torch.long)
        # (sent_idx, predicate, word_index) => span_label
        for sent_idx, sent_target_labels in enumerate(labels):
            for predicate, index, label in sent_target_labels:
                label_index = self.vocab.stoi[label]
                if label_index == 0:
                    continue
                label_tensor[sent_idx, predicate, index] = label_index
                label_set.add((sent_idx, predicate, index, label_index))
        return label_tensor.to(device), label_set



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
from functools import partial
from torch import nn
import ltp.nn


def MLP(input_size, hidden_size, dropout, **kwargs):
    activation_func = ltp.nn.ReLU
    for activation, args in kwargs.items():
        activation_func = partial(getattr(ltp.nn.activation, activation), **args)

    return nn.Sequential(
        nn.Linear(input_size, hidden_size),
        activation_func(),
        nn.Dropout(p=dropout)
    )



#! /usr/bin/env python
# -*- coding: utf-8 -*_
# Author: Yunlong Feng <ylfeng@ir.hit.edu.cn>
import inspect

from ltp.core import Registrable
from torch.optim import lr_scheduler
from transformers import get_linear_schedule_with_warmup, \
    get_constant_schedule, get_constant_schedule_with_warmup, \
    get_cosine_schedule_with_warmup, get_cosine_with_hard_restarts_schedule_with_warmup


class Scheduler(lr_scheduler._LRScheduler, metaclass=Registrable):
    """
    调度器
    """

    @classmethod
    def from_extra(cls, extra: dict, subcls=None):
        sig = inspect.signature(subcls)
        params = sig.parameters.keys()

        res = {}

        config = extra['config']
        init = config.get('init', {})
        if ('num_training_steps' in params) and ('num_training_steps' not in init):
            res['num_training_steps'] = extra['num_training_steps']

        if ('num_warmup_steps' in params) and ('num_warmup_steps' not in init) and ('warmup_proportion' in config):
            res['num_warmup_steps'] = extra['num_training_steps'] * config['warmup_proportion']
        return res


Scheduler.weak_register("LambdaLR", lr_scheduler.LambdaLR)
Scheduler.weak_register("CosineAnnealingLR", lr_scheduler.CosineAnnealingLR)
Scheduler.weak_register("ExponentialLR", lr_scheduler.ExponentialLR)
Scheduler.weak_register("MultiStepLR", lr_scheduler.MultiStepLR)
Scheduler.weak_register("ReduceLROnPlateau", lr_scheduler.ReduceLROnPlateau)
Scheduler.weak_register("StepLR", lr_scheduler.StepLR)

try:
    Scheduler.weak_register("CosineAnnealingWarmRestarts", lr_scheduler.CosineAnnealingWarmRestarts)
    Scheduler.weak_register("CyclicLR", lr_scheduler.CyclicLR)
except Exception as e:
    pass

Scheduler.weak_register("ConstantLR", get_constant_schedule)
Scheduler.weak_register("ConstantLRW", get_constant_schedule_with_warmup)
Scheduler.weak_register("CosineLRW", get_cosine_schedule_with_warmup)
Scheduler.weak_register("CosineHrLRW", get_cosine_with_hard_restarts_schedule_with_warmup)
Scheduler.weak_register("LinearLRW", get_linear_schedule_with_warmup)

from . import scheduler


class TemperatureScheduler(metaclass=Registrable):
    """
    蒸馏温度调度器
    """

    def __call__(self, logits_S, logits_T, base_temperature):
        return self.forward(logits_S, logits_T, base_temperature)

    def forward(self, logits_S, logits_T, base_temperature):
        raise NotImplementedError("未实现")


from . import temperature_scheduler

__all__ = ['Scheduler', 'TemperatureScheduler']
